import ctypes
import time

import numpy as np
np.set_printoptions(formatter={'int':hex})

from . import lib6502
from . import dtypes as d
from ...emulator import Emulator
from atrip import disassembler as disasm

import logging
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)


debug_frames = False


class Generic6502(Emulator):
    cpu = "6502"
    name = "6502"
    ui_name = "Generic 6502"

    serializable_attributes = []

    output_array_dtype = d.OUTPUT_DTYPE
    width = d.VIDEO_WIDTH
    height = d.VIDEO_HEIGHT

    low_level_interface = lib6502

    history_entry_dtype = disasm.dd.HISTORY_6502_DTYPE

    @property
    def current_cpu_status(self):
        pc, a, x, y, sp, p, _ = self.cpu_state
        dtype = d.STATESAV_DTYPE
        state = self.state_array[0:int(d.STATESAV_DTYPE.itemsize)].view(dtype=d.STATESAV_DTYPE)[0]
        # print("raw: %s" % self.raw_array[0:32])
        return "A=%02x X=%02x Y=%02x SP=%02x FLAGS=%02x PC=%04x cyc=%ld instr=%ld" % (a, x, y, sp, p, pc, self.status['cycles_since_power_on'][0], self.status['instructions_since_power_on'][0])

    @property
    def stack_pointer(self):
        return self.cpu_state[4]

    @stack_pointer.setter
    def stack_pointer(self, value):
        self.cpu_state[4] = value

    @property
    def program_counter(self):
        return self.cpu_state[0]

    @program_counter.setter
    def program_counter(self, value):
        self.cpu_state[0] = value

    def configure_emulator_defaults(self):
        lib6502.set_a2_emulation_mode(0)

    def generate_save_state_memory_blocks(self):
        cpu_offset = self.state_start_offset
        memory_offset = cpu_offset + d.CPU_DTYPE.itemsize
        segments = [
            (cpu_offset, d.CPU_DTYPE.itemsize, 0, "CPU Status"),
            (memory_offset, d.MAIN_MEMORY_SIZE, 0, "Main Memory"),
        ]
        self.save_state_memory_blocks.extend(segments)

    def calc_cpu_data_array(self):
        offset = self.state_start_offset
        dtype = d.CPU_DTYPE
        raw = self.raw_array[offset:offset + dtype.itemsize]
        print(("sizeof raw_array=%d raw=%d dtype=%d" % (len(self.raw_array), len(raw), dtype.itemsize)))
        dataview = raw.view(dtype=dtype)
        return dataview[0]

    def calc_main_memory_array(self):
        offset = self.state_start_offset + d.CPU_DTYPE.itemsize
        raw = self.raw_array[offset:offset + d.MAIN_MEMORY_SIZE]
        return raw

    def boot_from_raw(self, data, origin):
        # for now, simply copies data into main memory
        end = origin + len(data)
        log.debug(f"Copying data to memory: {origin:#04x}-{end:#04x}, data={data}")
        self.main_memory[origin:end] = data
        self.cpu_state[0] = origin
        lib6502.restore_state(self.output_raw)
        self.debug_state()
        self.last_boot_state = self.calc_current_state()

    # Emulator user input functions

    def coldstart(self):
        """Simulate an initial power-on startup.
        """
        lib6502.init_emulator(None)
        if self.last_boot_state is not None:
            lib6502.restore_state(self.last_boot_state)

    def warmstart(self):
        """Simulate a warm start; i.e. pressing the system reset button
        """
        lib6502.init_emulator(None)
        if self.last_boot_state is not None:
            lib6502.restore_state(self.last_boot_state)

    def keypress(self, ascii_char):
        self.send_char(ord(ascii_char))

    # Utility functions

    def load_disk(self, drive_num, pathname):
        lib6502.load_disk(drive_num, pathname)



# Standard library imports.
import sys
import os
import functools

# Major package imports.
import wx
import numpy as np

# Local imports.
from sawx.utils.command import Overlay
from ..utils.drawutil import get_bounds
from sawx.utils.sortutil import invert_rects, rect_ranges_to_indexes
#import sawx.framework.actions as fa
from ..commands.clipboard import PasteCommand, PasteRectCommand

from .char import CharViewer
from . import mouse_modes as m

import logging
log = logging.getLogger(__name__)


class MapViewer(CharViewer):
    name = "map"

    ui_name = "Map"

    valid_mouse_modes = [m.RectangularSelectMode, m.EyedropperMode, m.DrawMode, m.LineMode, m.SquareMode, m.FilledSquareMode]

    default_mouse_mode_cls = m.RectangularSelectMode

    def __init__(self, *args, **kwargs):
        CharViewer.__init__(self, *args, **kwargs)
        self.draw_pattern = np.zeros((1,), dtype=np.uint8)

    @property
    def window_title(self):
        return "Map: " + self.machine.font_renderer.name + ", " + self.machine.font_mapping.name + ", " + self.machine.color_standard_name

    ##### Selections

    def highlight_selected_ranges_in_segment(self, selected_ranges, segment):
        # This is default implementation which simply highlights everything
        # between the start/end values of each range. Other selection types
        # (rectangular selection) will need to be defined in the subclass
        segment.set_style_ranges_rect(selected_ranges, self.control.items_per_row, selected=True)

    ##### Clipboard & Copy/Paste

    @property
    def clipboard_data_format(self):
        return "numpy,columns"

    def get_paste_command(self, serialized_data, *args, **kwargs):
        print(serialized_data)
        print((serialized_data.source_data_format_name))
        if serialized_data.source_data_format_name == "numpy,columns":
            cmd_cls = PasteRectCommand
        cmd_cls = PasteCommand
        return cmd_cls(self.segment, serialized_data, *args, **kwargs)

    ##### Drawing pattern

    def set_draw_pattern(self, value):
        log.debug("new draw pattern: %s" % str(value))
        self.draw_pattern = np.asarray([value], dtype=np.uint8)



# generated file - recompile fstbt.s to make changes

# append jump target (lo, hi) and sector address list before saving to disk
fstbt = b"\x01\xa8\x8dP\xc0\x8dR\xc0\x8dW\xc0\xd0\t\xc6\xfe\x106\xa9\xd0\x8d\x0b\x08\xee\x1a\x08\xad\x86\x08\xc9\xc0\xf0d\xc9\xd0\xf0\xf2\xc9\xd1\xd0\x05\x8dT\xc0\xf0\xe9\xc9\xd2\xd0\x05\x8dU\xc0\xf0\xe0\xc9\xe0\x90\x0b)\x1f\x85\xfe\xa9$\x8d\x0b\x08\x10\xc6\x85'\xc8\xc0\x10\x90\t\xf0\x05 g\x08\xa8,\xa0\x01\x84=\xc8\xa5'\xf0\xba\x8a {\xf8\t\xc0H\xa9[H`\xe6A\x06@ o\x08\x18 t\x08\xe6@\xa5@)\x03*\x05+\xa8\xb9\x80\xc0\xa90L\xa8\xfcL"



import numpy as np

from . import errors
from .segments import SegmentData, EmptySegment, ObjSegment, RawSectorsSegment, DefaultSegment, SegmentedFileSegment, SegmentSaver
from .style_bits import get_style_bits
from .utils import *

import logging
log = logging.getLogger(__name__)
try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


def get_xex(segments, run_addr=None):
    segments_copy = [s for s in segments]  # don't affect the original list!
    main_segment = None
    sub_segments = []
    data_style = get_style_bits(data=True)
    total = 2
    runad = False
    for s in segments:
        total += 4 + len(s)
        if s.origin == 0x2e0:
            runad = True
    if not runad:
        words = np.empty([1], dtype='<u2')
        if run_addr:
            found = False
            for s in segments:
                if run_addr >= s.origin and run_addr < s.origin + len(s):
                    found = True
                    break
            if not found:
                raise errors.InvalidBinaryFile("Run address points outside data segments")
        else:
            run_addr = segments[0].origin
        words[0] = run_addr
        r = SegmentData(words.view(dtype=np.uint8))
        s = DefaultSegment(r, 0x2e0)
        segments_copy[0:0] = [s]
        total += 6
    bytes = np.zeros([total], dtype=np.uint8)
    rawdata = SegmentData(bytes)
    main_segment = DefaultSegment(rawdata)
    main_segment.data[0:2] = 0xff  # FFFF header
    main_segment.style[0:2] = data_style
    i = 2
    for s in segments_copy:
        # create new sub-segment inside new main segment that duplicates the
        # original segment's data/style
        new_s = DefaultSegment(rawdata[i:i+4+len(s)], s.origin)
        words = new_s.data[0:4].view(dtype='<u2')
        words[0] = s.origin
        words[1] = s.origin + len(s) - 1
        new_s.style[0:4] = data_style
        new_s.data[4:4+len(s)] = s[:]
        new_s.style[4:4+len(s)] = s.style[:]
        i += 4 + len(s)
        new_s.copy_user_data(s, 4)
        sub_segments.append(new_s)
    return main_segment, sub_segments


def get_bsave(segments, run_addr=None):
    # Apple 2 executables get executed at the first address loaded. If the
    # run_addr is not the first byte of the combined data, have to create a
    # new 3-byte segment with a "JMP run_addr" to go at the beginning
    origin = 100000000
    last = -1

    for s in segments:
        origin = min(origin, s.origin)
        last = max(last, s.origin + len(s))
        if _xd: log.debug("contiguous bytes needed: %04x - %04x" % (origin, last))
    if run_addr and run_addr != origin:
        # check if run_addr points to some location that has data
        found = False
        for s in segments:
            if run_addr >= s.origin and run_addr < s.origin + len(s):
                found = True
                break
        if not found:
            raise errors.InvalidBinaryFile("Run address points outside data segments")
        origin -= 3
        hi, lo = divmod(run_addr, 256)
        raw = SegmentData([0x4c, lo, hi])
        all_segments = [DefaultSegment(raw, origin=origin)]
        all_segments.extend(segments)
    else:
        all_segments = segments
    size = last - origin
    image = np.zeros([size + 4], dtype=np.uint8)
    words = image[0:4].view(dtype="<u2")  # always little endian
    words[0] = origin
    words[1] = size
    for s in all_segments:
        index = s.origin - origin + 4
        log.debug("setting data for $%04x - $%04x at index $%04x" % (s.origin, s.origin + len(s), index))
        image[index:index + len(s)] = s.data
    return image


def create_executable_file_data(filename, segments, run_addr=None):
    name = filename.lower()
    if name.endswith("xex"):
        base_segment, user_segments = get_xex(segments, run_addr)
        return base_segment.data, "XEX"
    elif name.endswith("bin") or name.endswith("bsave"):
        data = get_bsave(segments, run_addr)
        return data, "B"
    raise errors.UnsupportedContainer



import os

import numpy as np

from . import errors
from .segments import SegmentData, EmptySegment, ObjSegment
from .diskimages import DiskImageBase
from .utils import to_numpy

import logging
log = logging.getLogger(__name__)


class LocalFilesystemImage(DiskImageBase):
    def __init__(self, path):
        self.path = path

    def __str__(self, path="."):
        return f"Local filesystem output to: {self.path}"

    def save(self, filename=""):
        # This is to save the disk image containing the files on the disk image
        # to the local disk, which doesn't make sense when the disk image is
        # the filesystem.
        pass

    def find_dirent(self, name):
        path = os.path.join(self.path, name)
        if os.path.exists(path):
            return True
        raise errors.FileNotFound("%s not found on disk" % str(name))

    def write_file(self, name, filetype, data):
        path = os.path.join(self.path, name)
        with open(path, "wb") as fh:
            fh.write(data)

    def delete_file(self, name):
        pass


class LocalFilesystem():
    def __init__(self, path="."):
        self.image = LocalFilesystemImage(path)



import os

import pytest

import numpy as np

from atrip.compressors import bzip, dcm, gzip, lz4, lzma, unix_compress, zlib

compressors_and_decompressors = [
    bzip.BZipCompressor,
    gzip.GZipCompressor,
    lz4.LZ4Compressor,
    lzma.LZMACompressor,
    zlib.ZLibCompressor,
]

read_only_compressors = [
    dcm.DCMCompressor,
    unix_compress.UnixCompressor,
]

class TestCompressor:
    def setup(self):
        data = np.arange(4096, dtype=np.uint8)
        data[1::2] = np.repeat(np.arange(16, dtype=np.uint8), 128)
        data[::100] = 0x7f
        self.byte_data = data.tobytes()

    @pytest.mark.parametrize("compressor_cls", compressors_and_decompressors)
    def test_compressor(self, compressor_cls):
        compressor = compressor_cls()

        packed = compressor.calc_packed_data(self.byte_data)
        assert packed != self.byte_data
        unpacked = compressor.calc_unpacked_data(packed)

        print(len(self.byte_data), len(packed), len(unpacked))
        assert unpacked == self.byte_data

if __name__ == "__main__":
    t = TestCompressor()
    t.setup()
    t.test_compressor(gzip.GZipCompressor)
    t.test_compressor(bzip.BZipCompressor)
    t.test_compressor(lzma.LZMACompressor)



import numpy as np
try:
    import lz4.frame as lz4
except ImportError:
    lz4 = None

from .. import errors
from ..compressor import Compressor


class LZ4Compressor(Compressor):
    compression_algorithm = "lz4"

    def calc_unpacked_data(self, byte_data):
        if lz4 is None:
            raise errors.InvalidAlgorithm("lz4 module needed for .lz4 support")
        try:
            unpacked = lz4.decompress(bytes(byte_data))
        except RuntimeError as e:
            raise errors.InvalidAlgorithm(e)
        return unpacked

    def calc_packed_data(self, byte_data, media=None):
        if lz4 is None:
            raise errors.InvalidAlgorithm("lz4 module needed for .lz4 support")
        try:
            packed = lz4.compress(bytes(byte_data))
        except RuntimeError as e:
            raise errors.InvalidAlgorithm(e)
        return packed



class AtrError(RuntimeError):
    pass


class AtrWarning(RuntimeWarning):
    pass


class InvalidSegmentLength(AtrError):
    pass


class InvalidSegmentOrder(AtrError):
    pass


class InvalidDiskImage(AtrError):
    """ Disk image is not recognized by a parser.

    Usually a signal to try the next parser; this error doesn't propagate out
    to the user much.
    """
    pass


class UnsupportedDiskImage(AtrError):
    """ Disk image is recognized by a parser but it isn't supported yet.

    This error does propagate out to the user.
    """
    pass


class InvalidSegment(AtrError):
    pass


class ReadOnlyContainer(AtrError):
    pass


# Errors for compressors and other translators

class AlgorithmError(AtrError):
    pass

class UnsupportedAlgorithm(AlgorithmError):
    pass

class InvalidAlgorithm(AlgorithmError):
    pass


# Errors for archivers

class ArchiverError(AtrError):
    pass

class InvalidArchiver(ArchiverError):
    pass


# Errors when trying to determine media type

class MediaError(AtrError):
    pass


class InvalidMediaSize(MediaError):
    pass


class InvalidHeader(MediaError):
    pass


class InvalidSectorNumber(MediaError):
    pass


class UnsupportedSectorType(MediaError):
    pass


# Warnings when trying to determine media type, mostly used to identify
# possible media types that can't be fully specified

class MediaWarning(AtrWarning):
    pass


class PossibleCandidateMedia(MediaWarning):
    pass


# Errors when trying to determine filesystem. Raising one of these errors
# during filesystem detection will abort the process and report failure for
# that filesystem.

class FilesystemError(AtrError):
    pass


class IncompatibleMediaError(FilesystemError):
    pass


class InvalidDirent(FilesystemError):
    pass


class LastDirent(FilesystemError):
    pass


class NoSpaceInDirectory(FilesystemError):
    pass


class NotEnoughSpaceOnDisk(FilesystemError):
    pass


class FileNotFound(FilesystemError):
    pass


# Errors in files or structure of a file. These are separate from
# FilesystemError subclasses to indicate they aren't fatal when detecting
# filesystems

class FileError(AtrError):
    pass


class FileStructureError(FileError):
    pass


class FileNumberMismatchError164(FileError):
    pass


class ByteNotInFile166(FileError):
    pass


class InvalidBinaryFile(FileError):
    pass



from __future__ import print_function
from builtins import zip
from builtins import range
from builtins import object
import os

import numpy as np
import pytest

from atrip.container import Container
from atrip.segment import Segment
# from atrip import get_xex, interleave_segments, user_bit_mask, diff_bit_mask
from atrip import errors
from functools import reduce


def get_indexed(segment, num, scale):
    indexes = np.arange(num) * scale
    s = Segment(segment, indexes)
    return s, indexes

# class TestSegment1:
#     def setup(self):
#         self.segments = []
#         for i in range(8):
#             data = np.arange(1024, dtype=np.uint8) * i
#             r = SegmentData(data)
#             self.segments.append(Segment(r, i * 1024))

#     def test_xex(self):
#         items = [
#             [(0, 1, 2), 0],
#             ]
        
#         for indexes, stuff in items:
#             s = [self.segments[i] for i in indexes]
#             s[1].style[0:500] = diff_bit_mask
#             s[1].set_comment_at(0, "comment 0")
#             s[1].set_comment_at(10, "comment 10")
#             s[1].set_comment_at(100, "comment 100")
#             print(list(s[1].iter_comments_in_segment()))
#             with pytest.raises(errors.InvalidBinaryFile):
#                 seg, subseg = get_xex(s, 0xbeef)
#             seg, subseg = get_xex(s)
#             assert tuple(seg.data[0:2]) == (0xff, 0xff)
#             # 2 bytes for the ffff
#             # 4 bytes per segment for start, end address
#             # An extra segment has been inserted for the run address!
#             size = reduce(lambda a, b:a + len(b), subseg, 0)
#             assert len(seg) == 2 + size
#             print(id(s[1]), list(s[1].iter_comments_in_segment()))
#             print(id(subseg[2]), list(subseg[2].iter_comments_in_segment()))
#             for i, c in s[1].iter_comments_in_segment():
#                 assert c == subseg[2].get_comment(i + 4)
#             assert np.all(s[1].style[:] == subseg[2].style[4:])

#     def test_copy(self):
#         for s in self.segments:
#             d = s.rawdata
#             print("orig:", d.data.shape, d.is_indexed, d.data, id(d.data))
#             c = d.copy()
#             print("copy", c.data.shape, c.is_indexed, c.data, id(c.data))
#             assert c.data.shape == s.data.shape
#             assert id(c) != id(s)
#             assert np.all((c.data[:] - s.data[:]) == 0)
#             c.data[0:100] = 1
#             print(d.data)
#             print(c.data)
#             assert not np.all((c.data[:] - s.data[:]) == 0)


class TestIndexed:
    def setup(self):
        data = np.arange(4096, dtype=np.uint8)
        data[1::2] = np.repeat(np.arange(16, dtype=np.uint8), 128)
        self.container = Container(data)
        self.segment = Segment(self.container, 0, length=len(self.container))

    def test_offsets(self):
        assert np.array_equal(self.segment.container_offset, np.arange(len(self.container)))

    def test_subset(self):
        # get indexed, will result in every 3th byte
        s, indexes = get_indexed(self.segment, 256, 3)
        assert np.array_equal(s.container_offset, indexes)
        for i in range(len(indexes)):
            index_in_source = i * 3
            assert np.array_equal(s.container_offset[i], index_in_source)
            s[i] = 33
            assert s[i] == self.container[index_in_source]
            self.container[index_in_source] = 3
            assert s[i] == self.container[index_in_source]

        # get indexed into indexed, will result in every 9th byte
        s2, indexes2 = get_indexed(s, 64, 3)
        assert np.array_equal(s2.container_offset, indexes2 * 3)
        for i in range(len(indexes2)):
            index_in_source = i * 9
            assert np.array_equal(s2.container_offset[i], index_in_source)
            s2[i] = 99
            assert s2[i] == self.container[index_in_source]
            self.container[index_in_source] = 9
            assert s2[i] == self.container[index_in_source]

    # def test_indexed_sub(self):
    #     base = self.segment
    #     assert not base.rawdata.is_indexed
    #     raw = base.rawdata[512:1536]  # 1024 byte segment
    #     sub = Segment(raw, 512)
        
    #     assert not sub.rawdata.is_indexed
    #     for i in range(len(sub)):
    #         ri = sub.get_raw_index(i)
    #         assert ri == sub.origin + i
    #         assert sub[i] == base[ri]
    #     start, end = sub.byte_bounds_offset()
    #     assert start == 512
    #     assert end == 1536
        
    #     with pytest.raises(IndexError) as e:
    #         # attempt to get indexes to 1024 * 3... Index to big => fail!
    #         s, indexes = get_indexed(sub, 1024, 3)
        
    #     # try with elements up to 256 * 3
    #     s, indexes = get_indexed(sub, 256, 3)
    #     print(sub.data)
    #     print(indexes)
    #     print(s.data[:])
    #     assert s.rawdata.is_indexed
    #     for i in range(len(indexes)):
    #         ri = s.get_raw_index(i)
    #         print(ri, "base[ri]=%d" % base[ri], i, indexes[i], "s[i]=%d" % s[i])
    #         assert ri == sub.origin + indexes[i]
    #         assert s[i] == base[ri]
    #     start, end = s.byte_bounds_offset()
    #     assert start == 0
    #     assert end == len(base)
        
    #     # get indexed into indexed, will result in every 9th byte
    #     s2, indexes2 = get_indexed(s, 64, 3)
    #     assert s2.rawdata.is_indexed
    #     for i in range(len(indexes2)):
    #         assert s2.get_raw_index(i) == sub.origin + indexes2[i] * 3
    #     start, end = s.byte_bounds_offset()
    #     assert start == 0
    #     assert end == len(base)

    # def test_interleave(self):
    #     base = self.segment
    #     r1 = base.rawdata[512:1024]  # 512 byte segment
    #     s1 = Segment(r1, 512)
    #     r2 = base.rawdata[1024:1536]  # 512 byte segment
    #     s2 = Segment(r2, 1024)
        
    #     indexes1 = r1.get_indexes_from_base()
    #     verify1 = np.arange(512, 1024, dtype=np.uint32)
    #     assert np.array_equal(indexes1, verify1)
        
    #     indexes2 = r2.get_indexes_from_base()
    #     verify2 = np.arange(1024, 1536, dtype=np.uint32)
    #     assert np.array_equal(indexes2, verify2)
        
    #     s = interleave_segments([s1, s2], 2)
    #     a = np.empty(len(s1) + len(s2), dtype=np.uint8)
    #     a[0::4] = s1[0::2]
    #     a[1::4] = s1[1::2]
    #     a[2::4] = s2[0::2]
    #     a[3::4] = s2[1::2]
    #     print(list(s[:]))
    #     print(list(a[:]))
    #     print(s.rawdata.order)
    #     assert np.array_equal(s[:], a)
        
    #     s = interleave_segments([s1, s2], 4)
    #     a = np.empty(len(s1) + len(s2), dtype=np.uint8)
    #     a[0::8] = s1[0::4]
    #     a[1::8] = s1[1::4]
    #     a[2::8] = s1[2::4]
    #     a[3::8] = s1[3::4]
    #     a[4::8] = s2[0::4]
    #     a[5::8] = s2[1::4]
    #     a[6::8] = s2[2::4]
    #     a[7::8] = s2[3::4]
    #     assert np.array_equal(s[:], a)

    # def test_interleave_not_multiple(self):
    #     base = self.segment
    #     r1 = base.rawdata[512:1024]  # 512 byte segment
    #     s1 = Segment(r1, 512)
    #     r2 = base.rawdata[1024:1536]  # 512 byte segment
    #     s2 = Segment(r2, 1024)
        
    #     indexes1 = r1.get_indexes_from_base()
    #     verify1 = np.arange(512, 1024, dtype=np.uint32)
    #     assert np.array_equal(indexes1, verify1)
        
    #     indexes2 = r2.get_indexes_from_base()
    #     verify2 = np.arange(1024, 1536, dtype=np.uint32)
    #     assert np.array_equal(indexes2, verify2)
        
    #     s = interleave_segments([s1, s2], 3)

    #     # when interleave size isn't a multiple of the length, the final array
    #     # will reduce the size of the input array to force it to be a multiple.
    #     size = (len(s1) // 3) * 3
    #     assert len(s) == size * 2
    #     a = np.empty(len(s), dtype=np.uint8)
    #     a[0::6] = s1[0:size:3]
    #     a[1::6] = s1[1:size:3]
    #     a[2::6] = s1[2:size:3]
    #     a[3::6] = s2[0:size:3]
    #     a[4::6] = s2[1:size:3]
    #     a[5::6] = s2[2:size:3]
    #     assert np.array_equal(s[:], a)

    # def test_interleave_different_sizes(self):
    #     base = self.segment
    #     r1 = base.rawdata[512:768]  # 256 byte segment
    #     s1 = Segment(r1, 512)
    #     r2 = base.rawdata[1024:1536]  # 512 byte segment
    #     s2 = Segment(r2, 1024)
        
    #     indexes1 = r1.get_indexes_from_base()
    #     verify1 = np.arange(512, 768, dtype=np.uint32)
    #     assert np.array_equal(indexes1, verify1)
        
    #     indexes2 = r2.get_indexes_from_base()
    #     verify2 = np.arange(1024, 1536, dtype=np.uint32)
    #     assert np.array_equal(indexes2, verify2)
        
    #     s = interleave_segments([s1, s2], 3)

    #     # when interleave size isn't a multiple of the length, the final array
    #     # will reduce the size of the input array to force it to be a multiple.
    #     size = (min(len(s1), len(s2)) // 3) * 3
    #     assert size == (256 // 3) * 3
    #     assert len(s) == size * 2
    #     a = np.empty(len(s), dtype=np.uint8)
    #     a[0::6] = s1[0:size:3]
    #     a[1::6] = s1[1:size:3]
    #     a[2::6] = s1[2:size:3]
    #     a[3::6] = s2[0:size:3]
    #     a[4::6] = s2[1:size:3]
    #     a[5::6] = s2[2:size:3]
    #     assert np.array_equal(s[:], a)

    # def test_copy(self):
    #     s, indexes = get_indexed(self.segment, 1024, 3)
    #     c = s.rawdata.copy()
    #     print(c.data.shape, c.is_indexed)
    #     print(id(c.data.np_data), id(s.data.np_data))
    #     assert c.data.shape == s.data.shape
    #     assert id(c) != id(s)
    #     assert np.all((c.data[:] - s.data[:]) == 0)
    #     c.data[0:100] = 1
    #     assert not np.all((c.data[:] - s.data[:]) == 0)


# class TestComments:
#     def setup(self):
#         data = np.ones([4000], dtype=np.uint8)
#         r = SegmentData(data)
#         self.segment = Segment(r, 0)
#         self.sub_segment = Segment(r[2:202], 2)

#     def test_locations(self):
#         s = self.segment
#         s.set_comment([[4,5]], "test1")
#         s.set_comment([[40,50]], "test2")
#         s.set_style_ranges([[2,100]], comment=True)
#         s.set_style_ranges([[200, 299]], data=True)
#         for i in range(1,4):
#             for j in range(1, 4):
#                 # create some with overlapping regions, some without
#                 r = [500*j, 500*j + 200*i + 200]
#                 s.set_style_ranges([r], user=i)
#                 s.set_user_data([r], i, i*10 + j)
#         r = [100, 200]
#         s.set_style_ranges([r], user=4)
#         s.set_user_data([r], 4, 99)
#         r = [3100, 3200]
#         s.set_style_ranges([r], user=4)
#         s.set_user_data([r], 4, 99)

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#     def test_split_data_at_comment(self):
#         s = self.segment
#         s.set_style_ranges([[0,1000]], data=True)
#         for i in range(0, len(s), 25):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         print(orig[0:200])
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#         r = s2.get_entire_style_ranges([1], user=True)
#         print(r)
#         assert r == [((0, 23), 1), ((23, 48), 1), ((48, 73), 1), ((73, 98), 1), ((98, 123), 1), ((123, 148), 1), ((148, 173), 1), ((173, 198), 1), ((198, 200), 1)]

#     def test_split_data_at_comment2(self):
#         s = self.segment
#         start = 0
#         i = 0
#         for end in range(40, 1000, 40):
#             s.set_style_ranges([[start, end]], user=i)
#             start = end
#             i = (i + 1) % 8
#         for i in range(0, len(s), 25):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s2 = self.sub_segment
#         print(len(s2))
#         copy = s2.get_comment_locations()
#         print(copy)
#         # comments at 4 and 40 in the original means 2 and 38 in the copy
#         orig = s.get_comment_locations()
#         print(orig[0:200])
#         assert copy[2] == orig[4]
#         assert copy[28] == orig[38]

#         r = s2.get_entire_style_ranges([1], user=user_bit_mask)
#         print(r)
#         assert r == [((0, 38), 0), ((38, 48), 1), ((48, 73), 1), ((73, 78), 1), ((78, 118), 2), ((118, 158), 3), ((158, 198), 4), ((198, 200), 5)]

#     def test_restore_comments(self):
#         s = self.segment
#         s.set_style_ranges([[0,1000]], data=True)
#         for i in range(0, len(s), 5):
#             s.set_comment([[i,i+1]], "comment at %d" % i)

#         s1 = self.segment
#         print(len(s1))
#         indexes = [7,12]
#         r = s1.get_comment_restore_data([indexes])
#         print(r)
#         # force clear comments
#         s1.rawdata.extra.comments = {}
#         s1.style[indexes[0]:indexes[1]] = 0
#         r0 = s1.get_comment_restore_data([indexes])
#         print(r0)
#         for start, end, style, items in r0:
#             print(style)
#             assert np.all(style == 0)
#             for rawindex, comment in list(items.values()):
#                 assert not comment
#         s1.restore_comments(r)
#         r1 = s1.get_comment_restore_data([indexes])
#         print(r1)
#         for item1, item2 in zip(r, r1):
#             print(item1)
#             print(item2)
#             for a1, a2 in zip(item1, item2):
#                 print(a1, a2)
#                 if hasattr(a1, "shape"):
#                     assert np.all(a1 - a2 == 0)
#                 else:
#                     assert a1 == a2

#         s2 = self.sub_segment
#         print(len(s2))
#         indexes = [5,10]
#         r = s2.get_comment_restore_data([indexes])
#         print(r)
#         # force clear comments
#         s2.rawdata.extra.comments = {}
#         s2.style[indexes[0]:indexes[1]] = 0
#         r0 = s2.get_comment_restore_data([indexes])
#         print(r0)
#         for start, end, style, items in r0:
#             print(style)
#             assert np.all(style == 0)
#             for rawindex, comment in list(items.values()):
#                 assert not comment
#         s2.restore_comments(r)
#         r2 = s2.get_comment_restore_data([indexes])
#         print(r2)
#         for item1, item2 in zip(r, r2):
#             print(item1)
#             print(item2)
#             for a1, a2 in zip(item1, item2):
#                 print(a1, a2)
#                 if hasattr(a1, "shape"):
#                     assert np.all(a1 - a2 == 0)
#                 else:
#                     assert a1 == a2

#         for item1, item2 in zip(r1, r2):
#             print(item1)
#             print(item2)
#             # indexes won't be the same, but rawindexes and comments will
#             assert np.all(item1[2] - item2[2] == 0)
#             assert set(item1[3].values()) == set(item2[3].values())


# class TestResize:
#     def setup(self):
#         data = np.arange(4096, dtype=np.uint8)
#         data[1::2] = np.repeat(np.arange(16, dtype=np.uint8), 128)
#         r = SegmentData(data)
#         self.container = Segment(r, 0)
#         self.container.can_resize = True

#     def test_subset(self):
#         # check to see data a view of some rawdata will be the same when the
#         # rawdata is resized.
#         c = self.container
#         assert not c.rawdata.is_indexed
#         offset = 1000
#         s = Segment(c.rawdata[offset:offset + offset], 0)
#         assert not s.rawdata.is_indexed

#         # Check that the small view has the same data as its parent
#         for i in range(offset):
#             assert s[i] == c[i + offset]

#         # keep a copy of the old raw data of the subset
#         oldraw = s.rawdata.copy()
#         oldid = id(s.rawdata)

#         requested = 8192
#         oldsize, newsize = c.resize(requested)
#         assert newsize == requested
#         s.replace_data(c)  # s should point to the same offset in the resized data
#         assert id(s.rawdata) == oldid  # segment rawdata object should be same
#         assert id(oldraw.order) == id(s.rawdata.order)  # order the same
#         for i in range(offset):  # check values compared to parent
#             assert s[i] == c[i + offset]

#         # check for changes in parent/view reflected so we see that it's
#         # pointing to the same array in memory
#         newbase = c.rawdata
#         newsub = s.rawdata
#         print(c.rawdata.data[offset:offset+offset])
#         print(s.rawdata.data[:])
#         s.rawdata.data[:] = 111
#         print(c.rawdata.data[offset:offset+offset])
#         print(s.rawdata.data[:])
#         for i in range(offset):
#             assert s[i] == c[i + offset]

#     def test_indexed(self):
#         c = self.container
#         assert not c.rawdata.is_indexed
#         s, indexes = get_indexed(self.container, 1024, 3)
#         assert s.rawdata.is_indexed
#         for i in range(len(indexes)):
#             assert s.get_raw_index(i) == indexes[i]
#         requested = 8192
#         oldraw = s.rawdata.copy()
#         oldid = id(s.rawdata)
#         oldsize, newsize = c.resize(requested)
#         assert newsize == requested
#         s.replace_data(c)
#         assert id(s.rawdata) == oldid
#         assert id(oldraw.order) == id(s.rawdata.order)
#         for i in range(len(indexes)):
#             assert s.get_raw_index(i) == indexes[i]
#         newbase = c.rawdata
#         newsub = s.rawdata
#         print(c.rawdata.data)
#         print(s.rawdata.data[:])
#         s.rawdata.data[:] = 111
#         print(c.rawdata.data)
#         print(s.rawdata.data[:])
#         for i in range(len(indexes)):
#             assert c.rawdata.data[indexes[i]] == s.rawdata.data[i]



if __name__ == "__main__":
    t = TestIndexed()
    t.setup()
    t.test_subset()
    # t.test_indexed_sub()
    # t.test_interleave()
    # t = TestSegment1()
    # t.setup()
    # t.test_xex()
    # t.test_copy()
    # t = TestComments()
    # t.setup()
    # t.test_split_data_at_comment()
    # t.test_restore_comments()
    # t = TestResize()
    # t.setup()
    # t.test_subset()



import os
import sys

# Major package imports.
import numpy as np

# Enthought library imports.
from traits.api import HasTraits, Any, Bool, Int, Str, List, Dict, Enum, DictStrStr

from sawx import persistence
from sawx.events import EventHandler

# Local imports.
from . import fonts
from atrip.machines import atari8bit
from . import disasm
from . import memory_map
# from . import antic_renderers

import logging
log = logging.getLogger(__name__)


def restore_from_last_time():
    log.debug("Restoring Machine config")
    init_font_list()
    init_assemblers()

def init_font_list():
    try:
        Machine.font_list = persistence.get_bson_data("font_list")
    except IOError:
        # file not found
        Machine.font_list = []
    except ValueError:
        # bad JSON format
        Machine.font_list = []

def init_assemblers():
    if Machine.assembler_list is None:
        Machine.assembler_list = persistence.get_json_data("assembler_list", [])

        # With built-in MAC/65 compilation support, fix the list of default
        # assemblers to change the default assembler to MAC/65 if the user
        # hasn't made an alteration to the list.
        a = Machine.assembler_list
        if len(a) == 2 and a[0]['name'] == "cc65" and a[1]['name'] == "MAC/65":
            Machine.assembler_list = None

    if not Machine.assembler_list:
        Machine.assembler_list = Machine.guess_default_assemblers()

def remember_for_next_time():
    log.debug("Remembering Machine config")
    if Machine.font_list:
        persistence.save_bson_data("font_list", Machine.font_list)
    if Machine.assembler_list:
        persistence.save_json_data("assembler_list", Machine.assembler_list)


class Machine(HasTraits):
    """ Collection of classes that identify a machine: processor, display, etc.

    """

    # Traits

    name = Str

    mime_prefix = Str("application/octet-stream")

    disassembler = Any(transient=True)

    memory_map = Any(transient=True)

    antic_font_data = Any

    antic_color_registers = Any

    color_standard = Enum(0, 1)

    color_registers = Any(transient=True)

    bitmap_renderer = Any(transient=True)

    font_renderer = Any(transient=True)

    font_mapping = Any(transient=True)

    page_renderer = Any(transient=True)

    assembler = Any(transient=True)

    machine_metadata_changed_event = Any

    # Class attributes (not traits)

    font_list = None

    assembler_list = None

    text_font = None

    @classmethod
    def find_machine_by_mime(cls, mime, default_if_not_matched=False):
        for m in predefined['machine']:
            if mime.startswith(m.mime_prefix):
                return m
        if default_if_not_matched:
            return predefined['machine'][0]

    # Trait initializers

    def _machine_metadata_changed_event_default(self):
        return EventHandler(self, debug=True)

    def _name_default(self):
        return "Generic 6502"

    def _assembler_default(self):
        for asm in self.assembler_list:
            if 'system default' in asm:
                return asm
        return self.assembler_list[0]

    def _disassembler_default(self):
        return predefined['disassembler'][0]

    def _memory_map_default(self):
        return predefined['memory_map'][0]

    def _antic_font_data_default(self):
        return fonts.A8DefaultFont

    def _antic_color_registers_default(self):
        return list(colors.powerup_colors())

    def _color_standard_default(self):
        return 0  # NTSC

    def _color_registers_default(self):
        return self.get_color_registers()

    def _bitmap_renderer_default(self):
        return predefined['bitmap_renderer'][0]

    def _font_renderer_default(self):
        return predefined['font_renderer'][0]

    def _font_mapping_default(self):
        return predefined['font_mapping'][0]

    def _page_renderer_default(self):
        return predefined['page_renderer'][0]

    def __getstate__(self):
        state = super(Machine, self).__getstate__()
        try:
            del state["__traits_version__"]
        except KeyError:
            pass

        for name in self.all_trait_names():
            t = self.trait(name)
            if t.transient:
                if name in predefined:
                    value = getattr(self, name)
                    try:
                        state[name] = predefined[name].index(value)
                    except ValueError:
                        log.warning("No matching index for predefined[%s]; will use default upon load instead of %s" % (name, value))
            elif name == "memory_map":
                # convert into list of tuples so json won't mangle the integer
                # keys into strings
                value = getattr(self, name)
                state[name] = list(value.items())
            elif name == "antic_font_data":
                value = getattr(self, name)
                state[name] = value['uuid']
        return state

    def state_to_traits(self, state):
        for name in self.all_trait_names():
            if name not in state:
                # skip missing trait definitions
                continue
            log.debug("state_to_traits: %s, restoring from %s" % (name, state.get(name, None)))
            t = self.trait(name)
            if t.transient:
                if name in predefined:
                    index = state[name]
                    setattr(self, name, predefined[name][index])
            elif name == "memory_map":
                # convert into list of tuples so json won't mangle the integer
                # keys into strings
                value = state[name]
                setattr(self, name, {k:v for k,v in value})
            elif name == "antic_font_data":
                uuid = state[name]
                try:
                    font_data = fonts.builtin_font_data[uuid]
                except KeyError:
                    font_data = fonts.A8DefaultFont
                except TypeError:
                    # loading omnivore 1.0 format
                    log.error("found Omnivore 1.0 font segment")
                    font_data = fonts.A8DefaultFont
                setattr(self, name, font_data)
            else:
                try:
                    setattr(self, name, state[name])
                except KeyError:
                    pass

    def __setstate__(self, state):
        self.state_to_traits(state)
        # set up trait notifications
        self._init_trait_listeners()
    #

    def __eq__(self, other):
        return self.disassembler == other.disassembler and self.memory_map == other.memory_map

    # to be usable in dicts, py3 needs __hash__ defined if __eq__ is defined
    def __hash__(self):
        return id(self)

    def clone_machine(self):
        m = self.clone_traits()
        m.update_colors(m.antic_color_registers)
        return m

    def serialize_extra_to_dict(self, mdict):
        """Save extra metadata to a dict so that it can be serialized
        """
        s = self.__getstate__()
        mdict.update(s)

    def restore_extra_from_dict(self, e):
        self.state_to_traits(e)
        self.update_colors(self.antic_color_registers)
    #

    def update_colors(self, c):
        baseline = list(colors.powerup_colors())
        # need to operate on a copy of the colors to make sure we're not
        # changing some global value. Also force as python int so we're not
        # mixing numpy and python values.
        if len(c) == 5:
            baseline[4:9] = [int(i) for i in c]
        else:
            baseline[0:len(c)] = [int(i) for i in c]
        self.antic_color_registers = baseline
        self.color_registers = self.get_color_registers()
        print("COLOR REGISTERS", self.color_registers)
        self.machine_metadata_changed_event(True)

    def get_color_registers(self, antic_color_registers=None):
        color_converter = self.get_color_converter()
        registers = []
        if antic_color_registers is None:
            antic_color_registers = self.antic_color_registers
        for c in antic_color_registers:
            registers.append(color_converter(c))

        # make sure there are 16 registers for 4bpp modes
        i = len(registers)
        for i in range(len(registers), 16):
            registers.append((i*16, i*16, i*16))

        # Extend to 32 for dimmed copies of the 16 colors
        dim = []
        for r in registers:
            dim.append((r[0]/4 + 64, r[1]/4 + 64, r[2]/4 + 64))
        registers.extend(dim)
        return registers

    def get_color_converter(self):
        if self.color_standard == 0:
            return colors.gtia_ntsc_to_rgb
        return colors.gtia_pal_to_rgb

    def set_color_standard(self, std):
        self.color_standard = std
        self.update_colors(self.antic_color_registers)

    @property
    def color_standard_name(self):
        return "NTSC" if self.color_standard == 0 else "PAL"

    def set_bitmap_renderer(self, renderer):
        self.bitmap_renderer = renderer
        self.bitmap_shape_change_event = True

    def set_disassembler(self, disassembler):
        self.disassembler = disassembler
        self.disassembler_change_event = True

    def set_memory_map(self, memory_map):
        self.memory_map = memory_map
        self.disassembler_change_event = True

    def get_font_renderer_from_font_name(self, font_name):
        for r in predefined['font_renderer']:
            if r.name.startswith(font_name):
                return r
        return predefined['font_renderer'][0]

    def get_font_mapping_from_name(self, name):
        for r in predefined['font_mapping']:
            if r.name.startswith(name):
                return r
        return predefined['font_mapping'][0]

    def set_font_mapping(self, font_mapping=None):
        if font_mapping is None:
            font_mapping = self.font_mapping
        elif isinstance(font_mapping, str):
            font_mapping = self.get_font_mapping_from_name(font_mapping)
        self.font_mapping = font_mapping
        self.font_change_event = True

    def load_font(self, task, filename):
        try:
            fh = open(filename, 'rb')
            data = fh.read() + "\0"*1024
            data = data[0:1024]
            font = {
                'name': os.path.basename(filename),
                'data': data,
                'char_w': 8,
                'char_h': 8,
                }
            self.set_font(font)
            self.font_list.append(font)
            self.remember_fonts(task.window.application)
            task.machine_menu_changed = self
        except:
            raise

    def set_assembler(self, assembler):
        self.assembler = assembler
        self.disassembler_change_event = True

    def add_assembler(self, task, d):
        self.assembler_list.append(d)
        self.remember_assemblers(task.window.application)
        task.machine_menu_changed = self

    @classmethod
    def guess_default_assemblers(cls):
        asm_list = [
            {'comment char': ';',
             'origin': '*=',
             'data byte': '.byte',
             'data byte prefix': '$',
             'data byte separator': ', ',
             'name': "MAC/65",
             },
            {'comment char': ';',
             'origin': '.org',
             'data byte': '.byte',
             'data byte prefix': '$',
             'data byte separator': ', ',
             'name': "cc65",
             },
            {'comment char': ';',
             'origin': '.org',
             'data byte': '.byte',
             'data byte prefix': '$',
             'data byte separator': ', ',
             'name': "MADS",
             },
            {'comment char': ';',
             'origin': 'org',
             'data byte': 'hex',
             'data byte prefix': '',
             'data byte separator': '',
             'name': "Merlin",
             },
            ]
        return [dict(asm) for asm in asm_list]  # force a copy

    @classmethod
    def set_system_default_assembler(cls, task, asm):
        if 'system default' in asm:
            del asm['system default']
        for e in cls.assembler_list:
            if 'system default' in e:
                del e['system default']
        found = False
        for e in cls.assembler_list:
            if e == asm:
                asm['system default'] = True
                found = True
                break
        if not found:
            cls.assembler_list[0:0] = [asm]
        cls.remember_assemblers(task.window.application)
        task.machine_menu_changed = cls

    @classmethod
    def get_default_assembler(cls):
        found = None
        for e in cls.assembler_list:
            if 'system default' in e:
                found = e
        if not found:
            if cls.assembler_list:
                found = cls.assembler_list[0]
        return found

    @classmethod
    def set_assembler_list(cls, task, asms):
        default = None
        for e in asms:
            if 'system default' in e:
                if default:
                    del e['system default']
                else:
                    default = e
        cls.assembler_list = asms
        cls.remember_assemblers(task.window.application)
        task.machine_menu_changed = cls

    def verify_current_assembler(self):
        found = None
        for asm in self.assembler_list:
            if self.assembler == asm:
                found = asm
                break
        if not found:
            asm = self.get_default_assembler()
        self.set_assembler(asm)

    # Utility methods

    def get_disassembler(self, hex_lower, mnemonic_lower, document_memory_map=None, segment_memory_map=None):
        if not document_memory_map and not segment_memory_map:  # either None or empty dict
            mmap = self.memory_map()
        else:
            # Create a merged memory map with first the segment map then the
            # document map taking precedence over the machine memory map when
            # there are duplicates
            parent = memory_map.EmptyMemoryMap
            mmap = parent.__class__("CustomMemoryMap", (parent,), {"rmemmap": dict(self.memory_map.rmemmap), "wmemmap": dict(self.memory_map.wmemmap)})
            if document_memory_map:
                mmap.rmemmap.update(document_memory_map)
            if segment_memory_map:
                mmap.rmemmap.update(segment_memory_map)
        return self.disassembler(self.assembler, mmap, hex_lower, mnemonic_lower)

    def get_nop(self):
        return self.disassembler.get_nop()


Generic6502 = Machine(name="Generic 6502", disassembler=disasm.Basic6502Disassembler)

Atari800 = Machine(name="Atari 800", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Basic6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari800Undoc = Machine(name="Atari 800 (show undocumented opcodes)", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Undocumented6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari800Flagged = Machine(name="Atari 800 (highlight undocumented opcodes)", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Flagged6502Disassembler, memory_map=memory_map.Atari800MemoryMap)

Atari5200 = Machine(name="Atari 5200", mime_prefix="application/vnd.atari8bit", disassembler=disasm.Basic6502Disassembler, memory_map=memory_map.Atari5200MemoryMap)


predefined = {
    "machine": [
        Generic6502,
        Atari800,
        Atari800Undoc,
        Atari800Flagged,
        Atari5200,
        ],
    "memory_map": [
        memory_map.EmptyMemoryMap,
        memory_map.Atari800MemoryMap,
        memory_map.Atari5200MemoryMap,
        memory_map.Apple2MemoryMap,
        memory_map.KIM1MemoryMap,
        ],
    "disassembler": [
        disasm.Basic6502Disassembler,
        disasm.Undocumented6502Disassembler,
        disasm.Flagged6502Disassembler,
        disasm.Basic65C02Disassembler,
        disasm.Basic65816Disassembler,
        disasm.Basic6800Disassembler,
        disasm.Basic6809Disassembler,
        disasm.Basic6811Disassembler,
        disasm.Basic8051Disassembler,
        disasm.Basic8080Disassembler,
        disasm.BasicZ80Disassembler,
        ],
    # "bitmap_renderer": [
    #     antic_renderers.OneBitPerPixelB(),
    #     antic_renderers.OneBitPerPixelW(),
    #     antic_renderers.OneBitPerPixelPM1(),
    #     antic_renderers.OneBitPerPixelPM2(),
    #     antic_renderers.OneBitPerPixelPM4(),
    #     antic_renderers.OneBitPerPixelApple2Linear(),
    #     antic_renderers.ModeB(),
    #     antic_renderers.ModeC(),
    #     antic_renderers.ModeD(),
    #     antic_renderers.ModeE(),
    #     antic_renderers.GTIA9(),
    #     antic_renderers.GTIA10(),
    #     antic_renderers.GTIA11(),
    #     antic_renderers.TwoBitsPerPixel(),
    #     antic_renderers.FourBitsPerPixel(),
    #     antic_renderers.TwoBitPlanesLE(),
    #     antic_renderers.TwoBitPlanesLineLE(),
    #     antic_renderers.TwoBitPlanesBE(),
    #     antic_renderers.TwoBitPlanesLineBE(),
    #     antic_renderers.ThreeBitPlanesLE(),
    #     antic_renderers.ThreeBitPlanesLineLE(),
    #     antic_renderers.ThreeBitPlanesBE(),
    #     antic_renderers.ThreeBitPlanesLineBE(),
    #     antic_renderers.FourBitPlanesLE(),
    #     antic_renderers.FourBitPlanesLineLE(),
    #     antic_renderers.FourBitPlanesBE(),
    #     antic_renderers.FourBitPlanesLineBE(),
    #     ],
    # "font_renderer": [
    #     antic_renderers.Mode2(),
    #     antic_renderers.Mode4(),
    #     antic_renderers.Mode5(),
    #     antic_renderers.Mode6Upper(),
    #     antic_renderers.Mode6Lower(),
    #     antic_renderers.Mode7Upper(),
    #     antic_renderers.Mode7Lower(),
    #     antic_renderers.Apple2TextMode(),
    #     ],
    # "font_mapping": [
    #     antic_renderers.ATASCIIFontMapping(),
    #     antic_renderers.AnticFontMapping(),
    #     ],
    # "page_renderer": [
    #     antic_renderers.BytePerPixelMemoryMap(),
    #     ],
    }


# Apple2 = Machine(name="Apple ][", mime_prefix="application/vnd.apple2", disassembler=disasm.Basic65C02Disassembler, antic_font_data=fonts.A2DefaultFont, font_renderer=predefined['font_renderer'][7], font_mapping=predefined['font_mapping'][1], antic_color_registers=[4, 30, 68, 213, 15, 202, 148, 70, 0], memory_map=memory_map.Apple2MemoryMap)
# predefined['machine'].append(Apple2)



import numpy as np

from .. import errors
from ..file_type import FileType
from ..segment import Segment
from ..style_bits import get_style_bits

import logging
log = logging.getLogger(__name__)

try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


class RunAddressSegment(Segment):
    @property
    def run_address(self):
        return self[0] + 256 * self[1]


class InitAddressSegment(Segment):
    @property
    def init_address(self):
        return self[0] + 256 * self[1]


class ObjSegment(Segment):
    def __init__(self, parent, offset, name, length):
        Segment.__init__(self, parent, offset, name=name, length=length)

    def calc_segments(self):
        """Convenience method used by subclasses to create any sub-segments
        within this segment.

        """
        start = self[0] + 256 * self[1]
        end = self[2] + 256 * self[3]
        count = end - start + 1
        if start == 0x2e0:
            target = self[4] + 256 * self[5]
            s = RunAddressSegment(self, 4, origin=start, name=f"RUNAD: JMP ${target:04x}", length=2)
        elif start == 0x2e2:
            target = self[4] + 256 * self[5]
            s = InitAddressSegment(self, 4, origin=start, name=f"INITAD: JSR ${target:04x}", length=2)
        else:
            s = Segment(self, 4, name=f"[${start:04x}-${end:04x}]", origin=start, length=count)
        return [s]


class AtariObjectFile(FileType):
    """Parse a binary chunk into segments according to the Atari DOS object
    file format.
    
    Ref: http://www.atarimax.com/jindroush.atari.org/afmtexe.html
    """
    ui_name = "Atari 8-bit Object File"

    def calc_segments(self):
        segments = []
        size = len(self)
        b = self.data
        pos = 0
        first = True
        if _xd: log.debug("Initial parsing: size=%d" % size)
        while pos < size:
            if pos + 1 < size:
                header, = b[pos:pos+2].view(dtype='<u2')
            else:
                segments.append(Segment(self, pos, 0, "Incomplete Data", length=1))
                break
            if header == 0xffff:
                # Apparently 0xffff header can appear in any segment, not just
                # the first.  Regardless, it is ignored everywhere.
                pos += 2
                first = False
                continue
            elif first:
                raise errors.InvalidBinaryFile("Object file doesn't start with 0xffff")
            if _xd: log.debug("header parsing: header=0x%x" % header)
            if len(b[pos:pos + 4]) < 4:
                segments.append(Segment(self, 0, 0, "Short Segment Header", length=len(b[pos:pos + 4])))
                break
            start, end = b[pos:pos + 4].view(dtype='<u2')
            if end < start:
                raise errors.InvalidBinaryFile("Nonsensical start and end addresses")
            count = end - start + 1
            found = len(b[pos + 4:pos + 4 + count])
            if found < count:
                segments.append(Segment(self, 4 + count, pos, pos + 4, start, end, "Incomplete Data"))
                break
            s = ObjSegment(self, pos, f"Segment #{len(segments) + 1}", length=4 + count)
            segments.append(s)
            pos += 4 + count
        return segments



#!/usr/bin/env python
import os
import sys
import hashlib
from collections import defaultdict
import pprint

import slugify

def parse(filename, mime):
    data = open(filename, 'rb').read()
    h = hashlib.sha1(data).digest()
    name = os.path.basename(os.path.splitext(filename)[0])
    return len(data), h, name



if __name__ == '__main__':
    this_dir = os.getcwd()
    source_dir = os.path.join(this_dir, "..", "atrip/signatures/")
    mime_with_type = sys.argv[1]
    if ".type" in mime_with_type:
        mime, _ = mime_with_type.rsplit(".type", 1)
    else:
        mime = mime_with_type
    slug = slugify.slugify(mime.replace("application/x.", ""), separator="_")
    source = os.path.join(source_dir, slug) + ".py"
    try:
        with open(source, 'r') as fh:
            source_text = fh.read()
    except OSError:
        source_text = "sha1_signatures = {}"
    try:
        exec(source_text)
    except:
        raise
    print(sha1_signatures)
    new_signatures = defaultdict(dict)
    new_signatures.update(sha1_signatures)
    for filename in sys.argv[2:]:
        print(f"parsing {filename}")
        size, hash_string, name = parse(filename, mime)
        _, rem = divmod(size, 1024)
        if rem > 0:
            print(f"skipping {name}; not a ROM image size")
            continue
        print(f"{size} {hash_string} {name}")
        new_signatures[mime_with_type][hash_string] = name
    lines = []
    lines.append("sha1_signatures = {")
    for k,v in sorted(new_signatures.items()):
        lines.append(f"\"{k}\": {{")
        for h,n in sorted(v.items(), key=lambda a:(a[1], a[0])):
            lines.append(f"  {h}: {repr(n)},")
        lines.append("},")
    lines.append("}  # end sha1_signatures\n")

    print("\n".join(lines))
    with open(source, 'w') as fh:
        fh.write("\n".join(lines))


# cython: language_level=3
from libc.stdio cimport printf
from libc.string cimport strstr, strcasestr
import cython
import numpy as np
cimport numpy as np

from libudis.libudis cimport history_entry_t, emulator_history_t, parse_func_t, string_func_t, jmp_targets_t, label_info_t, label_storage_t

from atrip.disassemblers.dtypes import HISTORY_ENTRY_DTYPE

cdef extern:
    parse_func_t parser_map[]
    string_func_t stringifier_map[]
    parse_func_t find_parse_function(char *)
    string_func_t find_string_function(char *)


cdef char *hexdigits_lower = "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f404142434445464748494a4b4c4d4e4f505152535455565758595a5b5c5d5e5f606162636465666768696a6b6c6d6e6f707172737475767778797a7b7c7d7e7f808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeafb0b1b2b3b4b5b6b7b8b9babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecfd0d1d2d3d4d5d6d7d8d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7f8f9fafbfcfdfeff"
cdef char *hexdigits_upper = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F202122232425262728292A2B2C2D2E2F303132333435363738393A3B3C3D3E3F404142434445464748494A4B4C4D4E4F505152535455565758595A5B5C5D5E5F606162636465666768696A6B6C6D6E6F707172737475767778797A7B7C7D7E7F808182838485868788898A8B8C8D8E8F909192939495969798999A9B9C9D9E9FA0A1A2A3A4A5A6A7A8A9AAABACADAEAFB0B1B2B3B4B5B6B7B8B9BABBBCBDBEBFC0C1C2C3C4C5C6C7C8C9CACBCCCDCECFD0D1D2D3D4D5D6D7D8D9DADBDCDDDEDFE0E1E2E3E4E5E6E7E8E9EAEBECEDEEEFF0F1F2F3F4F5F6F7F8F9FAFBFCFDFEFF"


cdef class TextStorage:
    cdef int max_lines
    cdef public int num_lines
    cdef public np.ndarray label_info
    cdef label_info_t *label_info_data
    cdef public np.ndarray text_buffer
    cdef char *text_buffer_data
    cdef int text_buffer_size
    cdef char *text_ptr
    cdef np.uint32_t text_index

    def __init__(self, max_lines, buffer_size=None):
        if buffer_size is None:
            buffer_size = max_lines * 64
        self.alloc_arrays(max_lines, buffer_size)
        self.clear()

    def alloc_arrays(self, max_lines, buffer_size):
        self.label_info = np.zeros(max_lines * sizeof(label_info_t), dtype=np.uint8)
        self.label_info_data = <label_info_t *>self.label_info.data
        self.text_buffer_size = buffer_size
        self.text_buffer = np.zeros(buffer_size, dtype=np.uint8)
        self.text_buffer_data = <char *>self.text_buffer.data
        self.max_lines = max_lines

    def __len__(self):
        return self.num_lines

    def __contains__(self, index):
        return index > 0 and index < self.num_lines

    def __getitem__(self, index):
        cdef label_info_t *info
        cdef int i, start, count

        if isinstance(index, slice):
            lines = []
            for i in range(*index.indices(len(self))):
                info = &self.label_info_data[i]
                start = info.text_start_index
                count = info.line_length
                lines.append(self.text_buffer[start:start + count].tostring())
            return lines
        elif isinstance(index, int):
            i = index
            info = &self.label_info_data[i]
            start = info.text_start_index
            count = info.line_length
            return self.text_buffer[start:start + count].tostring()
        else:
            raise TypeError(f"index must be int or slice, not {type(index)}")

    def __iter__(self):
        cdef label_info_t *info
        cdef int i, start, count

        for i in range(self.num_lines):
            info = &self.label_info_data[i]
            start = info.text_start_index
            count = info.line_length
            yield self.text_buffer[start:start + count].tostring()

    cdef clear(self):
        self.num_lines = 0
        self.text_index = 0
        self.text_ptr = self.text_buffer_data

    cdef store(self, int count):
        cdef label_info_t *info
        cdef int i

        i = self.num_lines
        if i >= self.max_lines:
            print(f"CYTHON ERROR! Text buffer line count {self.max_lines} exceeded in TextStorage, attempted to save {i} lines")
        info = &self.label_info_data[i]
        info.line_length = count
        info.text_start_index = self.text_index
        self.text_index += count
        self.text_ptr += count
        self.num_lines += 1
        if self.text_index >= self.text_buffer_size:
            print(f"CYTHON ERROR! Text buffer size {self.text_buffer_size} exceeded in TextStorage, attempted to save {self.text_index} bytes")
        #print(f"stored line {i}, max={self.max_lines}, storage: index={self.text_index}, max={self.text_buffer_size}")

    def get_label_data_addr(self):
        return long(<long>&self.label_info_data[0])

    def get_text_storage_addr(self):
        return long(<long>&self.text_buffer_data[0])


cdef class StringifiedDisassembly:
    cdef public int origin
    cdef public int last_pc
    cdef public int start_index
    cdef public np.ndarray jmp_targets
    cdef jmp_targets_t *jmp_targets_data

    cdef public TextStorage disasm_text

    # internals
    cdef int mnemonic_case
    cdef char *hex_case

    def __init__(self, start_index, max_lines, jmp_targets, labels=None, mnemonic_lower=True, hex_lower=True):
        cdef char *text
        cdef long addr
        cdef np.ndarray arr

        self.start_index = start_index
        self.disasm_text = TextStorage(max_lines)
        self.mnemonic_case = 1 if mnemonic_lower else 0
        self.hex_case = hexdigits_lower if hex_lower else hexdigits_upper
        self.jmp_targets = jmp_targets
        self.jmp_targets_data = <jmp_targets_t *>self.jmp_targets.data

        if labels is not None:
            # print("LABELS:", labels.labels_raw, labels.labels_raw.data)
            # for i in range(0x70,0x100):
            #     try:
            #         print(f"label {i:02x} {labels[i]}")
            #     except KeyError:
            #         pass
            arr = labels.labels_raw
            self.jmp_targets_data.labels = <label_storage_t *>arr.data
        else:
            self.jmp_targets_data.labels = <label_storage_t *>0

        self.clear()
        # for disassembler_type in range(40):
        #     printf("stringifier[%d] = %lx\n", disassembler_type, stringifier_map[disassembler_type])

    def __len__(self):
        return self.disasm_text.num_lines

    def __getitem__(self, index):
        if isinstance(index, slice):
            lines = []
            for i in range(*index.indices(len(self))):
                lines.append(self.disasm_text[i])
            return lines
        elif isinstance(index, int):
            return self.disasm_text[index]

        else:
            raise TypeError(f"index must be int or slice, not {type(index)}")

    def __iter__(self):
        for i in range(len(self)):
            yield self.disasm_text[i]

    cdef clear(self):
        self.origin = 0
        self.last_pc = 0
        self.disasm_text.clear()

    cdef parse_history_entries(self, history_entry_t *h, int num_entries):
        if num_entries > self.disasm_text.max_lines:
            num_entries = self.disasm_text.max_lines
        cdef int count
        cdef string_func_t stringifier

        self.disasm_text.clear()
        while num_entries > 0:
            stringifier = stringifier_map[h.disassembler_type]
            # printf("disassembler: %d, stringifier: %lx\n", h.disassembler_type, stringifier)
            count = stringifier(h, self.disasm_text.text_ptr, self.hex_case, self.mnemonic_case, self.jmp_targets_data)
            self.disasm_text.store(count)
            num_entries -= 1
            h += 1

    cdef search(self, history_entry_t *h, int num_entries, search_bytes, int match_case):
        cdef char *search = search_bytes
        cdef char *text = self.disasm_text.text_ptr
        cdef char *found
        cdef int pc = h.pc
        matches = []
        # printf("Searching for: %s in %d lines\n", search, num_entries)
        while num_entries > 0:
            stringifier = stringifier_map[h.disassembler_type]
            count = stringifier(h, text, self.hex_case, self.mnemonic_case, self.jmp_targets_data)
            text[count] = 0

            if match_case:
                found = strstr(text, search)
            else:
                found = strcasestr(text, search)
            if found:
                matches.append((h.pc - pc, h.pc - pc + h.num_bytes))
                # printf("disassembler: %d, remaining: %d, line: %s, found: %s\n", h.disassembler_type, num_entries, text, found)
            num_entries -= 1
            h += 1

        # printf("Searched for: %s, found %d matches\n", search, len(matches))
        return matches



cdef class ParsedDisassembly:
    cdef history_entry_t *history_entries
    cdef public int max_entries
    cdef int entry_size
    cdef public np.ndarray raw_entries
    cdef int num_entries
    cdef public int origin
    cdef int last_pc
    cdef int current_pc
    cdef public np.ndarray jmp_targets
    cdef jmp_targets_t *jmp_targets_data
    cdef public int num_bytes
    cdef np.uint32_t *index_to_row_data
    cdef public np.ndarray index_to_row
    cdef int index_index
    cdef int max_text_lines

    def __init__(self, max_entries, origin, num_bytes):
        self.max_entries = max_entries
        self.entry_size = sizeof(history_entry_t)
        self.raw_entries = np.zeros((max_entries + 1) * self.entry_size, dtype=np.uint8)
        self.history_entries = <history_entry_t *>self.raw_entries.data
        self.num_entries = 0
        self.origin = origin
        self.last_pc = origin
        self.current_pc = origin
        self.jmp_targets = np.zeros(sizeof(jmp_targets_t), dtype=np.uint8)
        self.jmp_targets_data = <jmp_targets_t *>self.jmp_targets.data

        self.num_bytes = num_bytes
        self.index_to_row = np.zeros(num_bytes + 1, dtype=np.uint32)
        self.index_to_row_data = <np.uint32_t *>self.index_to_row.data
        self.index_index = 0

        self.max_text_lines = 256

    def __len__(self):
        return self.num_entries

    cdef parse_next(self, parse_func_t processor, unsigned char *src, int num_bytes):
        cdef history_entry_t *h = &self.history_entries[self.num_entries]
        cdef int last_pc = self.current_pc + num_bytes
        cdef np.uint32_t *index_list = self.index_to_row_data
        cdef int i
        while self.current_pc < last_pc and self.num_entries < self.max_entries:
            if num_bytes > 0:
                count = processor(h, src, self.current_pc, last_pc, self.jmp_targets_data)
                src += count
                num_bytes -= count
                self.current_pc += count
                for i in range(count):
                    index_list[self.index_index] = self.num_entries
                    self.index_index += 1
                self.num_entries += 1
                h += 1
            else:
                break
        if self.num_entries >= self.max_entries:
            print(f"CYTHON ERROR! ParsedDisassembly disassembly entries {self.max_entries} exceeded, attempted to save {self.num_entries}")
        if self.index_index > self.num_bytes:
            print(f"CYTHON ERROR! ParsedDisassembly index_to_row entries {self.num_bytes} exceeded, attempted to save {self.index_index}")

    cdef fix_offset_labels(self):
        # fast loop in C to check for references to addresses that are in the
        # middle of an instruction. If found, a discovered address is generated
        # at the first byte of the instruction
        cdef int pc = self.origin
        cdef int i = self.num_bytes
        cdef np.uint8_t *jmp_target = <np.uint8_t *>self.jmp_targets_data
        cdef np.uint32_t *index_to_row = self.index_to_row_data
        cdef np.uint8_t disassembler_type

        #print "pc=%04x, last=%04x, i=%04x" % (pc, pc + i, i)
        while i > 0:
            i -= 1
            old_label = jmp_target[(pc + i) & 0xffff]
            if old_label:
                #print "disasm_info: found label %04x, index_to_row[%04x]=%04x" % (pc + i, i, index_to_row[i])
                while index_to_row[i - 1] == index_to_row[i] and i > 1:
                    i -= 1
                #if labels[pc + i] == 0:
                #    print "  disasm_info: added label at %04x" % (pc + i)
                jmp_target[(pc + i) & 0xffff] = old_label

    def parse_test(self, np.uint8_t disasm_type, np.ndarray[np.uint8_t, ndim=1] src):
        cdef parse_func_t processor

        processor = parser_map[disasm_type]
        # printf("processor = %lx\n", processor)
        self.parse_next(processor, <unsigned char *>src.data, len(src))

    def stringify(self, int index, int num_lines_requested, labels=None, mnemonic_lower=True, hex_lower=True):
        cdef history_entry_t *h = &self.history_entries[index]
        output = StringifiedDisassembly(index, num_lines_requested, self.jmp_targets, labels, mnemonic_lower, hex_lower)
        output.parse_history_entries(h, num_lines_requested)
        return output

    def search(self, search_bytes, match_case=False, labels=None):
        cdef history_entry_t *h = &self.history_entries[0]
        output = StringifiedDisassembly(0, 100, self.jmp_targets, labels, not match_case, not match_case)
        return output.search(h, self.num_entries, search_bytes, match_case)


cdef int data_style = 0

cdef class DisassemblyConfig:
    cdef np.uint8_t c_split_comments[256]
    cdef parse_func_t segment_parsers[256]
    cdef np.uint8_t default_disasm_type

    def __init__(self, def_disasm_type=0, split_comments=[data_style]):
        cdef int i
        for i in range(256):
            self.c_split_comments[i] = 1 if i in split_comments else 0
        self.default_disasm_type = def_disasm_type

    def get_parser(self, num_entries, origin, num_bytes):
        # has to be a python function because it can be overridden in
        # subclasses
        return ParsedDisassembly(num_entries, origin, num_bytes)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def parse(self, segment, num_entries):
        cdef np.uint8_t s, t
        cdef int comment_bit_mask = 0x40

        src_copy = segment.data.tobytes()
        cdef np.uint8_t *src = <np.uint8_t *>src_copy
        style_copy = segment.style.tobytes()
        cdef np.uint8_t *c_style = <np.uint8_t *>style_copy
        disasm_type_copy = segment.disasm_type.tobytes()
        cdef np.uint8_t *c_disasm_type = <np.uint8_t *>disasm_type_copy
        cdef int num_bytes = len(src_copy)

        cdef int origin = segment.origin
        cdef int end_addr = origin + len(segment)
        cdef int pc = origin

        if num_bytes < 1:
            return self.get_parser(0, origin, 0)
        cdef ParsedDisassembly parsed = self.get_parser(num_entries, origin, num_bytes)

        cdef int first_index = 0
        cdef np.uint8_t current_disasm_type = c_disasm_type[0]
        if current_disasm_type > 127:
            current_disasm_type = self.default_disasm_type
        cdef int start_index
        cdef int end_index = 0
        cdef int count
        # print "CYTHON FAST_GET_ENTIRE", style_copy
        for end_index in range(1, num_bytes):
            s = c_style[end_index]
            t = c_disasm_type[end_index]
            if t > 127:
                t = self.default_disasm_type
            # print "%04x" % i, s, s2,
            if s & comment_bit_mask:
                if t == current_disasm_type and not self.c_split_comments[t]:
                    # print("same w/skippable comment")
                    continue
            elif t == current_disasm_type:
                # print "same"
                continue

            # process chuck here:
            start_index = first_index
            count = end_index - start_index
            # print("break here -> %x:%x = %s" % (start_index, end_index, current_disasm_type))
            processor = parser_map[current_disasm_type]
            parsed.parse_next(processor, src, count)
            src += count
            first_index = end_index
            current_disasm_type = t

        # process last chunk
        start_index = first_index
        end_index += 1  # i is last byte tested, need +1 to include it in the range
        count = end_index - start_index
        # print("final break here -> %x:%x = %s, count=%x" % (start_index, end_index, current_disasm_type, num_bytes))
        processor = parser_map[current_disasm_type]
        parsed.parse_next(processor, src, count)

        parsed.fix_offset_labels()
        # print("finished offset label generation")
        return parsed


cdef class StringifiedHistory:
    cdef public int origin
    cdef public int last_pc
    cdef public int start_index
    cdef public np.ndarray jmp_targets
    cdef jmp_targets_t *jmp_targets_data

    cdef public TextStorage history_text
    cdef public TextStorage result_text

    # internals
    cdef int mnemonic_case
    cdef char *hex_case

    def __init__(self, max_lines, labels=None, mnemonic_lower=True, hex_lower=True):
        cdef long addr
        cdef np.ndarray arr

        self.history_text = TextStorage(max_lines)
        self.result_text = TextStorage(max_lines)
        self.mnemonic_case = 1 if mnemonic_lower else 0
        self.hex_case = hexdigits_lower if hex_lower else hexdigits_upper
        self.clear()
        self.jmp_targets = np.zeros(sizeof(jmp_targets_t), dtype=np.uint8)
        self.jmp_targets_data = <jmp_targets_t *>self.jmp_targets.data

        if labels is not None:
            arr = labels.labels_raw
            self.jmp_targets_data.labels = <label_storage_t *>arr.data
        else:
            self.jmp_targets_data.labels = <label_storage_t *>0

    def __len__(self):
        return self.history_text.num_lines

    def __getitem__(self, index):
        if isinstance(index, slice):
            lines = []
            for i in range(*index.indices(len(self))):
                lines.append((self.history_text[i], self.result_text[i]
))
            return lines
        elif isinstance(index, int):
            return self.history_text[index], self.result_text[index]

        else:
            raise TypeError(f"index must be int or slice, not {type(index)}")

    def __iter__(self):
        for i in range(len(self)):
            yield self.history_text[i], self.result_text[i]

    cdef clear(self):
        self.history_text.clear()
        self.result_text.clear()

    cdef parse_history_entries(self, emulator_history_t *history, int index, int num_entries):
        if num_entries > self.history_text.max_lines:
            num_entries = self.history_text.max_lines
        cdef int count
        cdef int disassembler_type
        cdef string_func_t stringifier
        cdef history_entry_t *h

        self.clear()
        while num_entries > 0:
            h = &history.entries[index]
            disassembler_type = h.disassembler_type
            stringifier = stringifier_map[disassembler_type]
            # printf("disassembler: %d, stringifier: %lx\n", h.disassembler_type, stringifier)
            count = stringifier(h, self.history_text.text_ptr, self.hex_case, self.mnemonic_case, self.jmp_targets_data)
            self.history_text.store(count)

            if disassembler_type < 192:
                stringifier = stringifier_map[h.disassembler_type + 1]
            # printf("disassembler: %d, stringifier: %lx\n", h.disassembler_type, stringifier)
                count = stringifier(h, self.result_text.text_ptr, self.hex_case, self.mnemonic_case, self.jmp_targets_data)
            else:
                count = 0
            self.result_text.store(count)

            num_entries -= 1
            index += 1
            if index >= history.num_allocated_entries:
                index = 0

cdef class HistoryStorage:
    cdef public np.ndarray history_array
    cdef emulator_history_t *history
    cdef public np.ndarray entries

    def __init__(self, num_entries):
        self.history_array = np.zeros(sizeof(emulator_history_t), dtype=np.uint8)
        self.history = <emulator_history_t *>self.history_array.data
        # printf("libudis: __init__: history_storage: %lx\n", <int>self.history)
        self.history.num_allocated_entries = num_entries
        self.entries = np.zeros(num_entries, dtype=HISTORY_ENTRY_DTYPE)
        self.history.entries = <history_entry_t *>self.entries.data
        self.clear()

    def __len__(self):
        return self.history.num_entries

    def __getitem__(self, index):
        cdef int wrapped

        if isinstance(index, slice):
            lines = []
            for i in range(*index.indices(len(self))):
                wrapped = (i + self.history.first_entry_index) % self.history.num_allocated_entries
                lines.append(self.entries[wrapped])
            return lines
        elif isinstance(index, int):
            wrapped = (index + self.history.first_entry_index) % self.history.num_allocated_entries
            return self.entries[wrapped]
        else:
            raise TypeError(f"index must be int or slice, not {type(index)}")

    def __iter__(self):
        for i in range(self.num_lines):
            yield self.entries[i]

    @property
    def first_entry_index(self):
        return self.history.first_entry_index

    @property
    def latest_entry_index(self):
        return self.history.latest_entry_index

    @property
    def next_entry_index(self):
        cdef np.int32_t last = self.history.latest_entry_index
        cdef np.int32_t mod = self.history.num_allocated_entries
        return ((last + 1) % mod)

    @property
    def cumulative_count(self):
        return self.history.cumulative_count

    cdef clear(self):
        self.history.first_entry_index = 0
        self.history.latest_entry_index = -1
        self.history.num_entries = 0
        self.history.cumulative_count = 0

    def summary(self):
        start = self.history.first_entry_index
        last = self.history.latest_entry_index
        mod = self.history.num_allocated_entries
        num = self.history.num_entries
        print(f"number of entries: {mod}, used: {num}, cumulative {self.history.cumulative_count}")
        print(f"start of ring: {start}, latest: {last}")
        print("summary:", self)
        printf("libudis: history_storage: %lx\n", <long int>self.history)

    def debug_range(self, from_index):
        last = self.history.latest_entry_index
        mod = self.history.num_allocated_entries
        num = self.history.num_entries
        print(f"{mod} entries; {from_index} -> {last % mod}")
        cdef i = from_index % mod
        print(f"{i}: {self.entries[i]}")
        i = (from_index+1) % mod
        print(f"{i}: {self.entries[i]}")
        print("  ...")
        i = (last - 1) % mod
        print(f"{i}: {self.entries[i]}")
        i = last % mod
        print(f"{i}: {self.entries[i]}")

    def stringify(self, int index, int num_lines_requested, labels=None, mnemonic_lower=True, hex_lower=True):
        output = StringifiedHistory(num_lines_requested, labels, mnemonic_lower, hex_lower)
        index = (self.history.first_entry_index + index) % self.history.num_allocated_entries
        output.parse_history_entries(self.history, index, num_lines_requested)
        return output

    def stringify_to(self, StringifiedHistory output, int index, int num_lines_requested):
        index = (self.history.first_entry_index + index) % self.history.num_allocated_entries
        output.parse_history_entries(self.history, index, num_lines_requested)
        return output



import os
import sys

import wx
import numpy as np

from atrip import Container, Segment, style_bits
from atrip.machines.atari8bit import powerup_colors
from atrip.machines.atari8bit.jumpman import parser as ju
from atrip.machines.atari8bit.jumpman import playfield as jp

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg
from sawx.persistence import get_template

from ..ui.segment_grid import SegmentGridControl, SegmentTable
from ..ui.info_panels import InfoPanel
from ..arch.bitmap_renderers import BaseRenderer
from ..jumpman import savers as js
from .. import clipboard_helpers

from ..viewer import SegmentViewer
from ..jumpman import mouse_modes as jm
from ..commands import jumpman as jc
from .bitmap import BitmapGridControl, BitmapViewer, BitmapLineRenderer
from .info import NonCaretInfoViewer

import logging
log = logging.getLogger(__name__)
drawlog = logging.getLogger("refresh")


class JumpmanPlayfieldRenderer(BaseRenderer):
    """ Custom renderer instead of Antic Mode D renderer. Need to display
    highlighting on a per-pixel level which isn't possible with the Mode D
    renderer because the styling info is applied at the byte level and there
    are 4 pixels per byte in Mode D.

    So this renderer is one byte per pixel, but only uses the first 16 colors.
    It is mapped to the ANTIC color register order, so the first 4 colors are
    player colors, then the 5 playfield colors. A blank screen corresponds to
    the index value of 8, so the last playfield color.
    """
    name = "Jumpman 1 Byte Per Pixel"
    pixels_per_byte = 1
    bitplanes = 1

    def get_image(self, segment_viewer, bytes_per_row, nr, count, pixels, style):
        normal = style == 0
        highlight = (style & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        comment = (style & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        data = (style & style_bits.user_bit_mask) > 0
        match = (style & style_bits.match_bit_mask) == style_bits.match_bit_mask

        color_registers, h_colors, m_colors, c_colors, d_colors = self.get_colors(segment_viewer, list(range(32)))
        bitimage = np.empty((nr * bytes_per_row, 3), dtype=np.uint8)
        for i in range(32):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:] = segment_viewer.preferences.empty_background_color[0:3]
        return bitimage.reshape((nr, bytes_per_row, 3))


class JumpmanFrameRenderer(BitmapLineRenderer):
    def draw_grid(self, grid_control, dc, first_row, visible_rows, first_cell, visible_cells):
        model = grid_control.model
        first_col = self.cell_to_col(first_row, first_cell)
        last_cell = min(first_cell + visible_cells, self.num_cells)
        last_row = min(first_row + visible_rows, model.antic_lines)
        last_col = self.cell_to_col(last_row, last_cell - 1) + 1
        drawlog.debug("draw_grid: rows:%d,%d, cols:%d,%d" % (first_row, last_row, first_col, last_col))

        ul_rect = self.col_to_rect(first_row, first_col)
        lr_rect = self.col_to_rect(last_row - 1, last_col - 1)
        frame_rect = wx.Rect(ul_rect.x, ul_rect.y, lr_rect.x - ul_rect.x + lr_rect.width, lr_rect.y - ul_rect.y + lr_rect.height)

        bytes_per_row = model.items_per_row
        first_index = first_row * bytes_per_row
        last_index = last_row * bytes_per_row
        data = model.data[first_index:last_index]
        style = model.style[first_index:last_index]
        drawlog.debug("draw_grid: first_index:%d last_index:%d" % (first_index, last_index))

        array = grid_control.bitmap_renderer.get_image(grid_control.segment_viewer, bytes_per_row, last_row - first_row, last_index - first_index, data, style)
        width = array.shape[1]
        height = array.shape[0]
        drawlog.debug("Calculated image: %dx%d" % (width, height))
        if width > 0 and height > 0:
            # image returned will have the correct number of rows but will be
            # 160 pixels wide; need to crop to visible columns
            cropped = array[:,first_col:last_col,:]
            array = intscale(cropped, grid_control.zoom_h, grid_control.zoom_w)
            #print("bitmap: %d,%d,3 after scaling: %s" % (height, width, str(array.shape)))
            image = wx.Image(array.shape[1], array.shape[0])
            image.SetData(array.tobytes())
            bmp = wx.Bitmap(image)
            dc.SetClippingRegion(frame_rect)
            dc.DrawBitmap(bmp, frame_rect.x, frame_rect.y)


class JumpmanSegmentTable(cg.HexTable):
    invalid_playfield_model = None

    @classmethod
    def get_invalid_playfield_model(cls):
        if cls.invalid_playfield_model is None:
            c = Container([0], force_numpy_data=True)
            s = Segment(c)
            cls.invalid_playfield_model = jp.JumpmanPlayfieldModel(s)
        return cls.invalid_playfield_model

    def __init__(self, linked_base, bytes_per_row):
        s = linked_base.segment
        try:
            model = s.jumpman_playfield_model
        except AttributeError:
            model = self.get_invalid_playfield_model()
        self.model = model
        cg.HexTable.__init__(self, self.model.playfield, self.model.playfield.style, self.model.items_per_row, 0x7000)

    @property
    def segment(self):
        return self.linked_base.segment

    def get_label_at_index(self, index):
        return (index // 40)


class JumpmanGridControl(BitmapGridControl):
    default_table_cls = JumpmanSegmentTable

    keybinding_desc = {
        "handle_char_move_backspace": "Back",
        "handle_char_move_delete": "Delete",
    }

    @property
    def model(self):
        return self.table.model

    def set_viewer_defaults(self):
        self.items_per_row = 160
        self.zoom = 4
        self.want_col_header = False
        self.want_row_header = False

    def calc_line_renderer(self):
        if hasattr(self, 'segment_viewer'):
            return JumpmanFrameRenderer(self, self.segment_viewer)
        return SegmentGridControl.calc_line_renderer(self)

    def refresh_view(self, *args, **kwargs):
        drawlog.debug("refresh_view")
        self.model.draw_playfield(True)
        BitmapGridControl.refresh_view(self, *args, **kwargs)

    def keep_current_caret_on_screen(self, flags):
        pass

    def draw_carets(self, *args, **kwargs):
        pass

    ##### popup stuff

    def get_extra_actions(self):
        return []

    ##### selections

    def select_all(self, caret_handler):
        """ Selects the entire document
        """
        mouse_mode = self.mouse_mode
        if mouse_mode.can_paste:
            first = True
            for obj in self.model.level_builder.objects:
                print(("select_all: adding %s" % str(obj)))
                mouse_mode.add_to_selection(obj, not first)
                first = False
            if not first:
                self.refresh_view()

    def select_none(self, caret_handler):
        """ Clears any selection in the document
        """
        mouse_mode = self.mouse_mode
        if mouse_mode.can_paste:
            mouse_mode.objects = []
            self.refresh_view()

    def select_invert(self, caret_handler):
        """ Selects the entire document
        """
        mouse_mode = self.mouse_mode
        if mouse_mode.can_paste:
            first = True
            current = set(mouse_mode.objects)
            for obj in self.model.level_builder.objects:
                if obj not in current:
                    mouse_mode.add_to_selection(obj, not first)
                    first = False
            if not first:
                self.refresh_view()

    ##### Copy/paste

    def calc_clipboard_data_objs(self, focused_control):
        data_objs = []
        objects = self.mouse_mode.objects
        if len(objects) > 0:
            blob = clipboard_helpers.JumpmanBlob(objects)
            data_objs.append(blob.data_obj)
        return data_objs

    ##### Keyboard handling

    def handle_char_move_backspace(self, evt, flags):
        self.mouse_mode.delete_objects()

    handle_char_move_delete = handle_char_move_backspace


class JumpmanViewerToolbarMixin:
    viewer_category = "Jumpman"

    exclude_from_menubar = ["Media"]

    default_mouse_mode_cls = jm.AnticDSelectMode

    viewer_extra_toolbar_desc = [
        "jumpman_select_mode",
        "jumpman_draw_girder_mode",
        "jumpman_draw_double_girder_mode",
        "jumpman_draw_ladder_mode",
        "jumpman_draw_up_rope_mode",
        "jumpman_draw_down_rope_mode",
        "jumpman_erase_girder_mode",
        "jumpman_erase_ladder_mode",
        "jumpman_erase_rope_mode",
        "jumpman_draw_coin_mode",
        "jumpman_respawn_mode",
    ]


class JumpmanOtherViewerToolbarMixin(JumpmanViewerToolbarMixin):
    @property
    def current_level(self):
        for viewer in self.editor.viewers:
            if viewer.linked_base == self.linked_base and viewer.name == "jumpman":
                return viewer.current_level


class JumpmanControlMouseModeMixin:
    """Proxy class to send mouse mode changes to the main JumpmanViewer when
    the mouse mode gets changed on the auxiliary jumpman viewers.
    """
    def set_mouse_mode(self, mode_cls):
        # there shouldn't be multiple level viewers with the same linked base,
        # but if there are, make sure they are all set to the same tool.
        v = self.segment_viewer
        e = v.editor
        for viewer in e.viewers:
            if viewer.linked_base == v.linked_base and viewer.name == "jumpman":
                viewer.control.set_mouse_mode(mode_cls)

    def is_mouse_mode(self, mode_cls):
        # this is harder than the above if there are multiple level viewers.
        # Hopefully they will all be on the same tool, so just use the first
        # viewer we find.
        v = self.segment_viewer
        e = v.editor
        for viewer in e.viewers:
            if viewer.linked_base == v.linked_base and viewer.name == "jumpman":
                return viewer.control.is_mouse_mode(mode_cls)


class JumpmanViewer(JumpmanViewerToolbarMixin, BitmapViewer):
    name = "jumpman"

    ui_name = "Jumpman Level Editor"

    control_cls = JumpmanGridControl

    has_zoom = True

    zoom_text = "bitmap zoom factor"

    has_caret = False

    popup_menu_desc = [
        "jumpman_set_trigger",
        "jumpman_clear_trigger",
        None,
        ]
    popup_menu_desc.extend(BitmapViewer.popup_menu_desc)

    def __init__(self, *args, **kwargs):
        BitmapViewer.__init__(self, *args, **kwargs)

    ##### Properties

    @property
    def bitmap_renderer(self):
        if self._bitmap_renderer is None:
            self._bitmap_renderer = JumpmanPlayfieldRenderer()
        return self._bitmap_renderer

    @property
    def window_title(self):
        return "Jumpman Level Editor"

    @property
    def current_level(self):
        s = self.segment
        try:
            model = s.jumpman_playfield_model
        except AttributeError:
            model = JumpmanSegmentTable.get_invalid_playfield_model()
        return model

    @property
    def can_copy(self):
        return self.current_level.mouse_mode.can_copy

    def get_extra_segment_savers(self, segment):
        """Hook to provide additional ways to save the data based on this view
        of the data
        """
        return [js.JumpmanSaveAsATR, js.JumpmanSaveAsXEX]

    #### copy/paste

    supported_clipboard_data = [
        wx.CustomDataObject("jumpman,objects"),
        ]

    def calc_paste_command(self, clipboard_blob, *args, **kwargs):
        objects = clipboard_blob.data
        log.debug(f"calc_paste_command: pasting jumpman objects {objects}")
        for o in objects:
            # offset slightly so the pasted objects are seen
            o.x += 1
            o.y += 1
        self.current_level.save_objects(objects)

    def calc_delete_command(self):
        if self.can_copy:
            objects = self.current_level.mouse_mode.objects
            self.current_level.delete_objects(objects)

    #### update routines

    def set_event_handlers(self):
        super().set_event_handlers()
        self.linked_base.jumpman_trigger_selected_event += self.on_jumpman_trigger_selected

    def recalc_data_model(self):
        log.debug(f"recalc_data_model: segment={self.segment}")
        self.current_level.init_level_builder(self)
        self.antic_color_registers = self.current_level.level_colors

    def on_update_table_for_value_change(self, evt):
        pass

    def on_jumpman_trigger_selected(self, evt):
        new_trigger_root = evt[0]
        log.debug("on_jumpman_trigger_selected: %s selected=%s" % (self, str(new_trigger_root)))
        self.set_trigger_view(new_trigger_root)

    def on_update_table_for_style_change(self, evt):
        self.on_update_table_for_value_change(evt)

    def recalc_view(self):
        """Rebuild the entire UI after a document formatting (or other
        structural change) or loading a new document.
        """
        self.recalc_data_model()

    ##### Jumpman level construction

    def update_harvest_state(self):
        if not self.current_level.valid_level:
            return
        harvest_state = self.current_level.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_coins = len(harvest_state.coins)

        # FIXME: force redraw of level data here because it depends on the
        # level builder objects so it can count the number of items
        #self.level_data.refresh_view()
        #self.trigger_list.refresh_view()

    def set_trigger_view(self, trigger_root):
        level = self.current_level
        mouse_mode = level.mouse_mode
        if mouse_mode.can_paste:
            mouse_mode.objects = []
        level.set_trigger_root(trigger_root)
        self.can_erase_objects = trigger_root is not None
        self.control.refresh_view()

    #### command processing

    def save_changes(self, command_cls=None, add=False, keep_objects_selected=False):
        if command_cls is None:
            if add:
                command_cls = jc.CreateObjectCommand
            else:
                command_cls = jc.MoveObjectCommand
        log.debug(f"save_changes: command_cls: {command_cls}")
        level = self.current_level
        builder = level.level_builder
        source, level_addr, old_harvest_addr = level.get_level_addrs()
        level_data, harvest_addr, ropeladder_data, num_coins = builder.create_level_definition(level_addr, source[0x46], source[0x47])
        index = level_addr - source.origin
        ranges = [(0x18,0x2a), (0x3e,0x3f), (0x4e,0x50), (index,index + len(level_data))]
        pdata = np.empty([1], dtype=np.uint8)
        if level.coin_harvest_diff < 0:
            pdata[0] = num_coins
        else:
            pdata[0] = max(0, num_coins - level.coin_harvest_diff)
        hdata = np.empty([2], dtype=np.uint8)
        hdata.view(dtype="<u2")[0] = harvest_addr
        data = np.hstack([ropeladder_data, pdata, hdata, level_data])
        cmd = command_cls(source, ranges, data, keep_objects_selected)
        log.debug(f"save_changes: starting process_command: {command_cls}")
        self.editor.process_command(cmd)
        log.debug(f"save_changes: finished process_command: {cmd}")
        log.debug("saved changes, new objects=%s" % builder.objects)

    def save_assembly(self, source, ranges, data):
        cmd = jc.MoveObjectCommand(source, ranges, data)
        self.editor.process_command(cmd)


##### Trigger painting viewer

class TriggerList(JumpmanControlMouseModeMixin, wx.ListBox):
    """Trigger selector for choosing which trigger actions to edit
    """

    def __init__(self, parent, linked_base, mdict, viewer_cls, **kwargs):
        self.triggers = None

        wx.ListBox.__init__(self, parent, style=wx.LB_SINGLE, **kwargs)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LISTBOX, self.on_click)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.on_dclick)
        self.Bind(wx.EVT_MOTION, self.on_tooltip)

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 300
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def recalc_view(self):
        try:
            level = self.segment_viewer.segment.jumpman_playfield_model
        except (AttributeError, ValueError):
            self.SetItems([])
            self.triggers = [None]
        else:
            self.set_coins(level)

    def refresh_view(self):
        if self.IsShown():
            drawlog.debug("refreshing %s" % self)
            self.recalc_view()
            self.Refresh()
        else:
            drawlog.debug("skipping refresh of hidden %s" % self)

    def parse_coins(self, coins, items, triggers, indent=""):
        for coin in coins:
            label = indent + coin.trigger_str
            if coin.trigger_function:
                label += f", trigger function=${coin.trigger_function:x}"
            items.append(label)
            triggers.append(coin)
            children = []
            for p in coin.trigger_painting:
                if p.single:
                    children.append(p)
            if children:
                self.parse_coins(children, items, triggers, indent + "    ")

    def set_coins(self, level):
        items = ["Main Level"]
        triggers = [None]
        index = 1
        selected_index = 0
        state = level.screen_state
        if state is not None:
            self.parse_coins(state.sorted_coins, items, triggers)
            for index, trigger in enumerate(triggers):
                if trigger == level.trigger_root:
                    selected_index = index
        if len(items) != self.GetCount():
            self.SetItems(items)
        else:
            for i, item in enumerate(items):
                old = self.GetString(i)
                if old != item:
                    self.SetString(i, item)
        if selected_index < self.GetCount():
            self.SetSelection(selected_index)
        self.triggers = triggers

    def on_left_down(self, event):
        item = self.HitTest(event.GetPosition())
        if item >= 0:
            selected = self.GetSelection()
            if selected != item:
                wx.CallAfter(self.update_triggers_in_main_viewer, self.triggers[item])
        event.Skip()

    def on_click(self, event):
        # BUG: doesn't seem to get called when selecting a segment, using the
        # comments sidebar to jump to another segment, then attempting to
        # select that previous segment. This function never gets called in
        # that case, so I had to add the check on EVT_LEFT_DOWN
        is_selected = event.GetExtraLong()
        if is_selected:
            selected = event.GetSelection()
            wx.CallAfter(self.update_triggers_in_main_viewer, self.triggers[selected])
        event.Skip()

    def update_triggers_in_main_viewer(self, new_trigger_root):
        log.debug(f"new trigger root: {new_trigger_root}")
        self.segment_viewer.linked_base.jumpman_trigger_selected_event(new_trigger_root)

    def on_dclick(self, event):
        event.Skip()

    def on_tooltip(self, evt):
        pos = evt.GetPosition()
        selected = self.HitTest(pos)
        if selected >= 0:
            message = "coin #%d" % selected
        else:
            message = ""
        # FIXME: set message, maybe in title bar?
        evt.Skip()


class TriggerPaintingViewer(JumpmanOtherViewerToolbarMixin, NonCaretInfoViewer):
    name = "trigger_painting"

    ui_name = "Jumpman Trigger Painting"

    control_cls = TriggerList

    def recalc_data_model(self):
        pass

    def show_caret(self, control, index, bit):
        pass

    def on_update_table_for_value_change(self, evt):
        self.control.recalc_view()

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


class JumpmanInfoPanel(JumpmanControlMouseModeMixin, InfoPanel):
    fields = [
        ("text", "Level Number", 0x00, 2),
        ("atascii_gr2_0xc0", "Level Name", 0x3ec, 20),
        ("uint", "Points per Coin", 0x33, 2, 250),
        ("label", "# Coins", "num_coins", 42),
        ("coins_needed", "Coins Needed", 0x3e, 1, ["All", "All except 1", "All except 2", "All except 3", "All except 4"]),
        ("uint", "Bonus Value", 0x35, 2, 2500),
        ("dropdown", "Number of Bullets", 0x3d, 1, ["None", "1", "2", "3", "4"]),
        ("antic_colors", "Game Colors", 0x2a, 9),
        ("label", "# Columns with Ladders", "num_ladders", 12),
        ("label", "# Columns with Downropes", "num_downropes", 6),
    ]

    def is_valid_data(self):
        try:
            jm = self.segment_viewer.segment.jumpman_playfield_model
        except AttributeError:
            return False
        return jm.possible_jumpman_segment and bool(jm.level_builder.objects)


class LevelSummaryViewer(JumpmanOtherViewerToolbarMixin, NonCaretInfoViewer):
    name = "level_summary"

    ui_name = "Jumpman Level Summary"

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        control = JumpmanInfoPanel(parent, linked_base, size=(350, 150))
        return control

    def recalc_data_model(self):
        pass

    def show_caret(self, control, index, bit):
        pass

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


class JumpmanCustomCodePanel(JumpmanInfoPanel):
    fields = [
        ("multi_line_label", "Code Summary", "custom_code_info"),
        ("multi_line_label", "Action Vectors", "action_vector_info"),
        ("multi_line_label", "Coin Trigger Functions", "coin_trigger_info"),
        ("multi_line_label", "Other Labels", "other_label_info"),
    ]


class CustomCodeViewer(JumpmanOtherViewerToolbarMixin, NonCaretInfoViewer):
    name = "custom_code"

    ui_name = "Jumpman Custom Code"

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        control = JumpmanCustomCodePanel(parent, linked_base, size=(350, 150))
        return control

    def recalc_data_model(self):
        pass

    def show_caret(self, control, index, bit):
        pass

    def on_update_table_for_value_change(self, evt):
        self.recalc_view()

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


##### Class level utilities

def find_first_valid_segment_index(segments):
    """Find first valid jumpman level in the list of segments.

    This prefers segments that are exactly 0x800 bytes long, but will return
    longer segments that look like they contain level data if no segments of
    the exact length are found.
    """
    possible = []
    for i, segment in enumerate(segments):
        if ju.is_valid_level_segment(segment):
            possible.append(i)
            if len(segment) == 0x800:
                return i
    if possible:
        return possible[0]
    return 0



from atrip.memory_map import MemoryMap
from sawx.persistence import get_template, iter_templates

import logging
log = logging.getLogger(__name__)


machine_labels = {}

def load_memory_map(keyword):
    global machine_labels
    try:
        rwlabels = machine_labels[keyword]
    except KeyError:
        try:
            text = get_template(keyword)
        except OSError as e:
            try:
                text = get_template(keyword + ".labels")
            except OSError as e:
                log.error(f"Couldn't find memory map named '{keyword}'")
                return MemoryMap(keyword)
        rwlabels, rlabels, wlabels = MemoryMap.from_text(keyword, text)
        machine_labels[keyword] = rwlabels  # FIXME: temporary until get sets of memory map labels
    return rwlabels

available_memory_maps = {}

def calc_available_memory_maps():
    global available_memory_maps
    if not available_memory_maps:
        for template in iter_templates("labels"):
            available_memory_maps[template.keyword] = template
    return available_memory_maps



#!/usr/bin/env python
"""Sort cartridges by the result of guess_settings from the atari800
emulator. Running it like this:

    /home/rob/src/atari800/src/guess_settings -5200 -atari uncat/* >cart-info.txt

creates text output like this:

    uncat/Asteroids (USA) (Proto).a52: 5200    NTSC Atari   status: OK through 1000 frames (cart=19 'Standard 8 KB 5200')
    uncat/Asteroids (USA) (Proto).a52: 5200    NTSC Atari   status: FAIL (unidentified cartridge)
    uncat/Astro Chase (USA).a52: 5200    NTSC Atari   status: FAIL (BRK instruction) (cart=16 'One chip 16 KB 5200')
    uncat/Astro Chase (USA).a52: 5200    NTSC Atari   status: OK through 1000 frames (cart=6 'Two chip 16 KB 5200')
    uncat/Astro Chase (USA).a52: 5200    NTSC Atari   status: FAIL (unidentified cartridge)

Save that text file, and run this script with:

    python cart-sort dir/to/base/path cart-info.txt

and it will move carts to specific type directories based on the success or
failure shown in the output.
"""


import os
import sys


if __name__ == '__main__':
    base_dir = sys.argv[1]
    guess_settings = sys.argv[2]
    guess_settings = os.path.join(base_dir, guess_settings)
    try:
        with open(guess_settings, 'r') as fh:
            source_text = fh.read()
    except OSError:
        raise RuntimeError("Run guess_settings to generate list of working combinations")

    valid = {}
    for line in source_text.splitlines():
        if "status: OK" not in line:
            continue
        filename, extra = line.split(":", 1)
        if "cart=" not in extra:
            print(f"Error: {filename}: not a cartridge")
            continue
        _, num = extra.split("cart=", 1)
        num, _ = num.split(" ", 1)
        if filename not in valid:
            print(f"{filename}: cart-type = {num}")
            valid[filename] = num
        else:
            print(f"Warning: {filename}: cart-type = {num} already identified as cart-type = {valid[filename]}")
            continue

        subdir = os.path.join(base_dir, f"type{num}")
        try:
            os.mkdir(subdir)
        except FileExistsError:
            pass
        src = os.path.join(base_dir, filename)
        dest = os.path.join(subdir, os.path.basename(filename))
        print(f"{src} -> {dest}")
        os.rename(src, dest)





"""Byte modification actions
"""
import os
import sys

import wx

from sawx.ui import dialogs

from ..action import ViewerAction, ViewerListAction, ViewerRadioListAction
from ..commands import disasm as dc
from atrip.disassembler import cpu_id_to_name, valid_cpu_ids

import logging
log = logging.getLogger(__name__)


class disasm_type(ViewerListAction):
    def calc_enabled(self, action_key):
       return self.viewer.control.caret_handler.has_selection

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return item[1]

    def calc_list_items(self):
        d = [(id, name) for id, name in cpu_id_to_name.items() if id not in valid_cpu_ids]
        d.sort()
        d[0:0] = [(128, "Code")]
        return d

    def perform(self, action_key):
        item = self.get_item(action_key)
        disasm_type = item[0]
        e = self.editor
        ranges = self.viewer.control.get_selected_ranges()
        cmd = dc.SetDisasmCommand(e.segment, ranges, disasm_type)
        e.process_command(cmd)



import hashlib
import uuid
from collections import defaultdict

import numpy as np
import wx

from sawx.persistence import iter_templates

from atrip.machines import atari8bit

from . import colors

import logging
log = logging.getLogger(__name__)

# Font is a dict (easily serializable with JSON) with the following attributes:
#    data: string containing font data
#    name: human readable name
#    x_bits: number of bits to display
#    y_bytes: number of bytes per character
#
# template:
# Font = {
#    'data_bytes': data_bytes,
#    'name':"Default Atari Font",
#    'char_w': 8,
#    'char_h': 8,
#    'uuid': uuid.UUID(bytes=hashlib.md5(data_bytes).digest()),
#    }


class AnticFont(object):
    def __init__(self, segment_viewer, font_data, font_renderer, playfield_colors, reverse=False):
        self.char_w = font_renderer.char_bit_width
        self.char_h = font_renderer.char_bit_height
        self.scale_w = font_renderer.scale_width
        self.scale_h = font_renderer.scale_height
        self.uuid = font_data['uuid']

        self.set_colors(segment_viewer, playfield_colors)
        self.set_fonts(segment_viewer, font_data, font_renderer, reverse)

    def set_colors(self, segment_viewer, playfield_colors):
        fg, bg = atari8bit.gr0_colors(playfield_colors)
        conv = segment_viewer.color_standard
        fg = conv(fg)
        bg = conv(bg)
        self.normal_gr0_colors = [fg, bg]
        prefs = segment_viewer.preferences
        self.highlight_gr0_colors = colors.get_blended_color_registers(self.normal_gr0_colors, prefs.highlight_background_color)
        self.match_gr0_colors = colors.get_blended_color_registers(self.normal_gr0_colors, prefs.match_background_color)
        self.comment_gr0_colors = colors.get_blended_color_registers(self.normal_gr0_colors, prefs.comment_background_color)
        self.data_gr0_colors = colors.get_dimmed_color_registers(self.normal_gr0_colors, prefs.background_color, prefs.data_background_color)

    def set_fonts(self, segment_viewer, font_data, font_renderer, reverse):
        if 'np_data' in font_data:
            data = font_data['np_data']
        else:
            data = np.fromstring(font_data['data_bytes'], dtype=np.uint8)
        self.font_data = font_data

        m = segment_viewer
        self.normal_font = font_renderer.get_font(data, m.color_registers, self.normal_gr0_colors, reverse)

        prefs = segment_viewer.preferences
        h_colors = colors.get_blended_color_registers(m.color_registers, prefs.highlight_background_color)
        self.highlight_font = font_renderer.get_font(data, h_colors, self.highlight_gr0_colors, reverse)

        d_colors = colors.get_dimmed_color_registers(m.color_registers, prefs.background_color, prefs.data_background_color)
        self.data_font = font_renderer.get_font(data, d_colors, self.data_gr0_colors, reverse)

        m_colors = colors.get_blended_color_registers(m.color_registers, prefs.match_background_color)
        self.match_font = font_renderer.get_font(data, m_colors, self.match_gr0_colors, reverse)

        c_colors = colors.get_blended_color_registers(m.color_registers, prefs.comment_background_color)
        self.comment_font = font_renderer.get_font(data, c_colors, self.comment_gr0_colors, reverse)

    def get_height(self, zoom):
        return self.char_h * self.scale_h * zoom

    def get_image(self, char_index, zoom, highlight=False):
        f = self.highlight_font if highlight else self.normal_font
        array = f[char_index]
        w = self.char_w
        h = self.char_h
        image = wx.Image(w, h)
        image.SetData(array.tobytes())
        w *= self.scale_w * zoom
        h *= self.scale_h * zoom
        image.Rescale(w, h)
        bmp = wx.Bitmap(image)
        return bmp


class FontListBox(wx.VListBox):
    def __init__(self, parent, font_uuids, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.SetItems(font_uuids)

    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            c = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            c = self.GetForegroundColour()
        dc.SetFont(self.GetFont())
        dc.SetTextForeground(c)
        dc.DrawLabel(self.get_item_text(n), rect,
                     wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)

    def OnMeasureItem(self, n):
        height = 0
        for line in self.get_item_text(n).split('\n'):
            w, h = self.GetTextExtent(line)
            height += h
        return height + 5

    def SetItems(self, items):
        self._items = [(valid_fonts[uuid], uuid) for uuid in items]
        self._items.sort(key=lambda a:a[0]['name'])
        self.SetItemCount(len(items))

    def EnsureVisible(self, num):
        print("first: %s" % self.GetFirstVisibleLine())
        print("last: %s" % self.GetLastVisibleLine())
        self.ScrollToLine(num)

    def GetCount(self):
        return self.GetItemCount()

    def get_font(self, index):
        font, uuid = self._items[index]
        return font

    def get_item_text(self, index):
        return self.get_font(index)['name']


class FontListDialog(wx.Dialog):
    """Simple dialog to return a font from the font group
    """
    border = 5

    def __init__(self, parent, font_uuids, default, title=None, callback=None):
        if title is None:
            title = "Choose Font"
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        self.list = FontListBox(self, font_uuids, size=(300,400))
        try:
            index = font_uuids.index(default)
        except ValueError:
            index = 0
        self.list.SetSelection(index)

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.list, 1, wx.EXPAND)
        sizer.Add(btnsizer, 0, wx.EXPAND, 0)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

        self.list.Bind(wx.EVT_LISTBOX, self.on_click)
        self.list.Bind(wx.EVT_LISTBOX_DCLICK, self.on_click)
        self.callback = callback

    def on_click(self, evt):
        if evt.IsSelection() or True:
            index = evt.GetSelection()
            font = self.list.get_font(index)
            log.debug(f"index={index} font={font['name']}")
            if self.callback:
                self.callback(font)

    def show_and_get_font(self):
        result = self.ShowModal()
        if result == wx.ID_OK:
            index = self.list.GetSelection()
            font = self.list.get_font(index)
        else:
            font = None
        self.Destroy()
        return font


class FontGroupDialog(FontListDialog):
    def __init__(self, parent, font_group_name, default=None, callback=None):
        font_uuids = font_groups[font_group_name]
        FontListDialog.__init__(self, parent, font_uuids, default, f"Choose Font From {font_group_name}", callback)


def prompt_for_font_from_group(parent, font_group_name, callback=None):
    d = FontGroupDialog(parent, font_group_name, callback=callback)
    return d.show_and_get_font()


font_list = []

font_groups = {}

valid_fonts = {}

def restore_from_last_time():
    global font_list, font_groups, valid_fonts

    log.debug("Restoring known fonts")
    groups = defaultdict(list)
    for template in iter_templates("font"):
        template['data_bytes'] = template.data_bytes
        valid_fonts[template['uuid']] = template
        log.debug(f"found font {template} at {template.data_file_path}")
        if "font_group" in template:
            groups[template['font_group']].append(template['uuid'])
        else:
            font_list.append(template)
    font_groups = groups
    font_list.sort()


def remember_for_next_time():
    log.debug("Remembering fonts")



import os
import io
import hashlib
import weakref

import numpy as np

from . import errors
from .utils import to_numpy, to_numpy_list, uuid
from .container import guess_container, Container, ContainerHeader
from .compressor import guess_compressor_list, compress_in_reverse_order, Uncompressed
from .archiver import Archiver, find_container_items_in_archive, PlainFileArchiver
from .filesystem import Dirent

import logging
log = logging.getLogger(__name__)


class Collection:
    """Parent object for single file archive collection of multiple disk images

    Instances of this class are not `Segment`s, but instead hold a list of
    `Container`s, each of which will have its own media type and filesystem
    structure, independent of the other disk images.
    """
    ui_name = "Collection"

    def __init__(self, pathname, data=None, session=None, container=None, guess=True):
        self.pathname = pathname
        self.name = ""
        self.containers = []
        self.decompression_order = [Uncompressed]
        self._uuid_map = None
        self.archiver = None
        if container is None:
            if data is None:
                data = open(pathname, 'rb').read()
                log.debug(f"Collection.__init__: {pathname}: read {len(data)} bytes")
            self.unarchive(data, session)
        else:
            self.archiver = PlainFileArchiver()
            self.add_container(container, pathname, guess=guess)

    @property
    def verbose_info(self):
        lines = []
        name = self.name or self.pathname
        lines.append(f"{name}: {self}")
        for c in self.containers:
            lines.append(c.container_info("    "))
        return "\n".join(lines)

    @property
    def basename(self):
        return os.path.basename(self.pathname)

    @property
    def mime_type(self):
        return "application/octet-stream"

    @property
    def uuid_map(self):
        if self._uuid_map is None:
            self._uuid_map = {}
            for segment in self.iter_segments():
                self._uuid_map[segment.uuid] = weakref.ref(segment)
            log.debug(f"created uuid map: {self._uuid_map}")
        return self._uuid_map

    #### dunder methods

    def __str__(self):
        desc = ""
        if len(self) > 1:
            desc = f"{len(self)} item "
        desc += f"{str(self.archiver)}"
        return desc

    def __len__(self):
        return np.alen(self.containers)

    def __getitem__(self, index):
        """Returns segment given the uuid, raising KeyError if not found
        """
        ref = self.uuid_map[index]
        segment = ref()
        return segment

    #### decompression

    def add_container(self, container, pathname, guess=True):
        container.pathname = pathname
        if guess:
            container.guess_media_type()
            container.guess_filesystem()
        self.containers.append(container)
        container.name = f"D{len(self.containers)}"
        log.info(f"container: {container}")

    def unarchive(self, byte_data, session=None):
        """Attempt to unpack `byte_data` using this archive unpacker.

        Calls `find_containers` to loop through each container found. The order
        listed here will be the order returned by the subclass; no sorting is
        done here.
        """
        decompressed_archive_byte_data, decompression_list = guess_compressor_list(byte_data)
        self.archiver, item_data_list = find_container_items_in_archive(self.pathname, decompressed_archive_byte_data)
        if session is not None:
            self.restore_session(session, item_data_list)
        elif not self.archiver.supports_multiple_containers:
            item_pathname, item_data = item_data_list[0]
            container = Container(item_data, decompression_list)
            self.add_container(container, item_pathname)
        else:
            self.decompression_order = decompression_list
            for item_pathname, item_data in item_data_list:
                log.info(f"container size: {len(item_data)}")
                container = guess_container(item_data)
                self.add_container(container, item_pathname)
        self._uuid_map = None

    def iter_archive(self, basename, byte_data):
        """Return a list of `Container` objects for each item in the archive.

        If the data is not recognized by this subclass, raise an
        InvalidContainer exception. This signals to the caller that a different
        container type should be tried.

        If the data is recognized by this subclass but the unpacking algorithm
        is not implemented, raise an UnsupportedContainer exception. This is
        different than the InvalidContainer exception because it indicates that
        the data was indeed recognized by this subclass (despite not being
        unpacked) and checking further containers is not necessary.
        """
        return [basename, byte_data]

    #### compression and save

    def save(self, pathname=None, skip_missing_compressors=False):
        """Save the collection.

        If pathname is None, will attempt to overwrite the file that used to
        load the collection.

        Can raise InvalidAlgorithm if one of the compressors is read-only
        (i.e. can only decompress data). However, if `skip_missing_compressors`
        is True, no error will be raised and compression will take place
        ignoring any compressors that can't compress data.
        """
        if pathname is None:
            pathname = self.pathname
        compressed_bytes = self.calc_compressed_data(skip_missing_compressors)
        with open(pathname, 'wb') as fh:
            fh.write(compressed_bytes)

    def calc_compressed_data(self, skip_missing_compressors=False):
        fh = io.BytesIO()
        self.save_in_archive(fh, skip_missing_compressors)
        archived_bytes = fh.getvalue()
        compressed_bytes = compress_in_reverse_order(archived_bytes, self.decompression_order)
        return compressed_bytes

    def save_in_archive(self, fh, skip_missing_compressors=False):
        """Pack each container into the archive
        """
        return self.archiver.pack_data(fh, self.containers, skip_missing_compressors)

    #### iterators

    def iter_segments(self):
        for container in self.containers:
            yield from container.iter_segments()

    def iter_menu(self):
        for container in self.containers:
            yield (container, 0)
            yield from container.iter_menu(1)

    def iter_dirents(self):
        for container in self.containers:
            yield from container.iter_dirents()

    #### search utilities

    def find_interesting_segment_to_edit(self):
        """Find the the first segment that might be "interesting" that an
        editor can use as an initial segment display.

        Largely depends on the filesystem to locate something interesting.

        Returns: None if nothing is super interesting
        """
        segment = None
        for container in self.containers:
            if container.filesystem is not None:
                segment = container.filesystem.find_interesting_segment()
                if segment is not None:
                    break
        return segment

    def find_boot_media(self):
        """Find the first bootable media in the collection.

        Should usually be in the first container, but will continue looking
        through containers if not in the first.
        """
        for container in self.containers:
            if container.media is not None:
                return container.media
        log.warning(f"No bootable media found in {self}, attempting to find an interesting segment")
        for s in self.iter_segments():
            if isinstance(s, ContainerHeader):
                continue
            return s
        else:
            try:
                return self.containers[0]
            except IndexError:
                raise errors.MediaError("find_boot_media: no containers, so no booting!")

    def find_dirent(self, filename, match_case=False):
        try:
            disk_input, pathname = filename.split(":", 1)
        except ValueError:
            disk_input = "D1"
            pathname = filename
        disk = disk_input.upper()
        try:
            if disk.startswith("D"):
                disk = disk[1:]
            disk_num = int(disk)
        except ValueError:
            raise errors.FileNotFound(f"Invalid disk specifier {disk_input}")
        try:
            container = self.containers[disk_num - 1]  # disk numbers 1 based
        except IndexError:
            raise errors.FileNotFound(f"Disk {disk_input} doesn't exist")
        return container.find_dirent(pathname, match_case)

    def find_uuid(self, uuid):
        try:
            segment = self[uuid]
        except KeyError:
            raise errors.InvalidSegment(f"No segment in any disk with uuid={uuid}")
        return segment

    #### session

    def serialize_session(self, e):
        """Save session information to a dict so that it can be serialized.

        Note that only the structure of the containers is saved, not the byte
        data. Styling and disassembler info is saved, however.
        """
        e["pathname"] = self.pathname
        e["name"] = self.name
        e["archiver"] = self.archiver
        e["containers"] = self.containers

    def restore_session(self, e, item_data_list=None):
        # log.debug("restoring sesssion data: %s" % str(e))
        self.pathname = e["pathname"]
        self.name = e["name"]
        self.archiver = e["archiver"]

        if item_data_list is None:
            item_data_list = [(c.pathname, c._data) for c in self.containers]

        # unless restoring a session over top of existing containers, the
        # containers will have been initialized with zeros from the jsonpickle
        # restore. The arrays will be the correct size, so all we have to do is
        # copy the actual data into the containers.
        self.containers = e["containers"]
        for c, (item_pathname, item_data) in zip(self.containers, item_data_list):
            c._data[:] = np.frombuffer(item_data, dtype=np.uint8)
            c.pathname = item_pathname



from ..assembler import Assembler, AssemblerResult
from .libmac65 import mac65_assemble

class MAC65(Assembler):
    name = "mac65"
    ui_name = "MAC/65 (ATasm)"
    cpu = "6502"

    comment_char = ";"
    origin = "*="
    data_byte = ".byte"
    data_byte_prefix = "$"
    data_byte_separator = ", "

    def __init__(self, verbose=False):
        Assembler.__init__(self, verbose)
        self.current_parser = self.null_parser

    def assemble(self, source):
        if isinstance(source, str):
            source = source.encode("utf-8")
        result = AssemblerResult()
        errors, text = mac65_assemble(source)
        errors = errors.strip()
        if errors:
            result.errors = [errors]
        self.current_parser = self.null_parser
        if text:
            self.parse(result, text)
        return result

    def null_parser(self, result, line, cleanup=False):
        pass

    def source_parser(self, result, line, cleanup=False):
        if cleanup:
            if result.first_addr is None:
                return
            if self.verbose: print("Code block: %x-%x" % (result.first_addr, result.last_addr))
            result.segments.append((result.first_addr, result.last_addr, result.current_bytes))
            result.first_addr = None
            result.last_addr = None
            result.current_bytes = []
            return

        lineno, addr, data, text = line[0:5], line[6:10], line[12:30], line[30:]
        addr = int(addr, 16)
        b = [int(a,16) for a in data.split()]
        #print hex(index), b
        if b:
            count = len(b)
            if result.first_addr is None:
                result.first_addr = result.last_addr = addr
            elif addr != result.last_addr:
                if self.verbose: print("Code block: %x-%x" % (result.first_addr, result.last_addr))
                result.segments.append((result.first_addr, result.last_addr, result.current_bytes))
                result.first_addr = result.last_addr = addr
                result.current_bytes = []

            result.current_bytes.extend(b)
            result.last_addr += count

            try:
                code_type = self.code_type_at_line[lineno]
            except KeyError:
                directive = text.strip()
                if len(directive) == len(text):
                    # label
                    print("LABEL!!!", lineno, text)
                    if len(text) > 0:
                        _, directive = text.split(None, 1)
                    else:
                        directive = ".byte"
                if directive.startswith("."):
                    if directive == ".word":
                        code_type = "w"
                    else:
                        code_type = "b"
                else:
                    code_type = "code"
                self.code_type_at_line[lineno] = code_type
            for i in range(count):
                result.addr_type_code[addr + i] = code_type
            if self.verbose: print(f"line: {addr:04x}, {text}")

    def equates_parser(self, result, line, cleanup=False):
        if cleanup:
            return
        symbol, addr = line.split(": ")
        if symbol[0] == "*":
            result.transitory_equates[symbol[1:].lower()] = int(addr, 16)
        else:
            result.equates[symbol.lower()] = int(addr, 16)

    def symbol_parser(self, result, line, cleanup=False):
        if cleanup:
            return
        symbol, addr = line.split(": ")
        result.add_label(symbol, addr)

    def parse(self, result, text):
        self.code_type_at_line = {}
        for line in text.splitlines():
            line = line.strip()
            if not line:
                continue
            if self.verbose: print("parsing:", line)
            if line.startswith("Source:"):
                self.current_parser(result, None, cleanup=True)
                self.current_parser = self.source_parser
            elif line == "Equates:":
                self.current_parser(result, None, cleanup=True)
                self.current_parser = self.equates_parser
            elif line == "Symbol table:":
                self.current_parser(result, None, cleanup=True)
                self.current_parser = self.symbol_parser
            else:
                self.current_parser(result, line)
        result.generate_data_info()



import sys
import unittest

from sawx import events


def mp_do_nothing(sender, *args):
    # Does nothing
    pass


class BasicEventHandlerTest(unittest.TestCase):
    def test_EventHandler(self):
        self.assertRaises(TypeError, events.EventHandler)
        self.assertIsInstance(events.EventHandler(None), events.EventHandler)
        self.assertIsInstance(events.EventHandler(132), events.EventHandler)
        self.assertIsInstance(events.EventHandler("Test"), events.EventHandler)

        ev = events.EventHandler(None)
        self.assertEqual(ev.sender, None)
        ev = events.EventHandler("Test")
        self.assertEqual(ev.sender, "Test")
        self.assertEqual(len(ev), 0)
        self.assertEqual(len(ev.callbacks), 0)

    def test_EventHandler_add__iadd__(self):
        ev = events.EventHandler(None)

        def doadd(ev, cb):
            ev += cb

        def callback():
            pass

        self.assertRaises(TypeError, doadd, ev, None)
        self.assertRaises(TypeError, doadd, ev, "Test")
        self.assertRaises(TypeError, doadd, ev, 1234)

        self.assertEqual(len(ev), 0)
        ev += callback
        self.assertEqual(len(ev), 1)
        for x in range(4):
            ev += callback
        self.assertEqual(len(ev), 5)

        self.assertRaises(TypeError, ev.add, None)
        self.assertRaises(TypeError, ev.add, "Test")
        self.assertRaises(TypeError, ev.add, 1234)

        self.assertEqual(len(ev), 5)
        ev.add(callback)
        self.assertEqual(len(ev), 6)
        for x in range(4):
            ev.add(callback)
        self.assertEqual(len(ev), 10)

    def test_EventHandler_remove__isub__(self):
        ev = events.EventHandler(None)

        def doremove(ev, cb):
            ev -= cb

        def callback():
            pass

        for x in range(10):
            ev += callback
        self.assertEqual(len(ev), 10)

        self.assertRaises(TypeError, ev.remove)
        for invval in ("Test", None, 1234, self.assertEqual):
            self.assertRaises(ValueError, ev.remove, invval)
            self.assertRaises(ValueError, doremove, ev, invval)
        self.assertEqual(len(ev), 10)
        ev.remove(callback)
        self.assertEqual(len(ev), 9)
        ev -= callback
        self.assertEqual(len(ev), 8)
        for x in range(3):
            ev.remove(callback)
            ev -= callback
        self.assertEqual(len(ev), 2)

    def test_EventHandler__call__(self):
        ev = events.EventHandler("Test")
        testsum = []

        def callback(evt):
            self.assertEqual(evt.sender, "Test")
            sumval = evt[0]
            sumval.append(1)

        for x in range(10):
            ev += callback
        self.assertEqual(len(ev), 10)
        results = ev(testsum)
        self.assertEqual(len(testsum), 10)
        for v in testsum:
            self.assertEqual(v, 1)
        self.assertEqual(len(results), 10)
        for v in results:
            self.assertIsNone(v)

    def test_EventHandler__call__WithKeywords(self):
        ev = events.EventHandler("Test")
        testsum = []

        def callback(evt):
            self.assertEqual(evt.sender, "Test")
            evt.testsum.append(1)

        for x in range(10):
            ev += callback
        self.assertEqual(len(ev), 10)
        results = ev(testsum=testsum)
        self.assertEqual(len(testsum), 10)
        for v in testsum:
            self.assertEqual(v, 1)
        self.assertEqual(len(results), 10)
        for v in results:
            self.assertIsNone(v)


class TestWeakrefEventHandler:
    def setup(self):
        pass

    def test_weakref_removal(self):
        ev = events.EventHandler("weakref")
        testsum = []

        def callback_function(evt):
            assert evt.sender == "weakref"
            sumval = evt[0]
            sumval.append(1)

        class receiver_object:
            def callback(self, evt):
                assert evt.sender == "weakref"
                sumval = evt[0]
                sumval.append(1)

        r = receiver_object()

        ev += r.callback
        ev += callback_function

        results = ev(testsum)
        assert len(testsum) == 2

        del r
        testsum = []

        results = ev(testsum)
        assert len(testsum) == 1

    def test_weakref_removal_with_keywords(self):
        ev = events.EventHandler("weakref")
        testsum = []

        def callback_function(evt):
            assert evt.sender == "weakref"
            evt.testsum.append(1)

        class receiver_object:
            def callback(self, evt):
                assert evt.sender == "weakref"
                evt.testsum.append(1)

        r = receiver_object()

        ev += r.callback
        ev += callback_function

        results = ev(testsum=testsum)
        assert len(testsum) == 2

        del r
        testsum = []

        results = ev(testsum=testsum)
        assert len(testsum) == 1





if __name__ == '__main__':
    sys.exit(unittest.main())
